###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""BookkeepingManager service is the front-end to the Bookkeeping database."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities.Decorators import deprecated
from DIRAC.ConfigurationSystem.Client.PathFinder import getServiceSection
from DIRAC.ConfigurationSystem.Client.Helpers import cfgPath
from DIRAC.ConfigurationSystem.Client.Config import gConfig


from LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB import OracleBookkeepingDB
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.XMLFilesReaderManager import XMLFilesReaderManager
from LHCbDIRAC.BookkeepingSystem.Client import JEncoder
from LHCbDIRAC.BookkeepingSystem.DB.Utilities import checkEnoughBKArguments

__RCSID__ = "$Id$"

# pylint: disable=invalid-name


default = 'ALL'


class BookkeepingManagerHandler(RequestHandler):

  """Bookkeeping Service class.

  It serves the requests made the users by using the BookkeepingClient.
  """

  @classmethod
  def initializeHandler(cls, serviceInfoDict):
    """Initializes the variables used to identify queries, which are not
    containing enough conditions."""

    cls.bkkDB = OracleBookkeepingDB()
    cls.xmlReader = XMLFilesReaderManager()
    cls.__eventTypeCache = {}

    bkkSection = getServiceSection("Bookkeeping/BookkeepingManager")
    if not bkkSection:
      cls.email = 'lhcb-bookkeeping@cern.ch'
      cls.forceExecution = False
    else:
      cls.email = gConfig.getValue(cfgPath(bkkSection, 'Email'), 'lhcb-bookkeeping@cern.ch')
      cls.forceExecution = gConfig.getValue(cfgPath(bkkSection, 'ForceExecution'), False)
    gLogger.info("Email used to track queries: %s forceExecution" % cls.email, cls.forceExecution)
    return S_OK()

  #############################################################################
  types_sendXMLBookkeepingReport = [six.string_types]

  def export_sendXMLBookkeepingReport(self, xml):
    """This method is used to upload an xml report which is produced after when
    the job successfully finished. The input parameter 'xml' is a string which
    contains various information (metadata) about the finished job in the Grid
    in an XML format.

    :param str xml: bookkeeping report
    """
    try:
      retVal = self.xmlReader.readXMLfromString(xml)
      if not retVal['OK']:
        self.log.error("Issue reading XML", retVal['Message'])
        return retVal
      if retVal['Value'] == '':
        return S_OK("The send bookkeeping finished successfully!")
      return retVal
    except Exception as x:
      errorMsg = "XML processing error"
      self.log.exception(errorMsg, lException=x)
      return S_ERROR(errorMsg)

  #############################################################################
  types_getAvailableSteps = [dict]

  @classmethod
  def export_getAvailableSteps(cls, in_dict):
    """It returns all the available steps which corresponds to a given
    conditions.

    The in_dict contains the following conditions: StartDate, StepId,
    InputFileTypes, OutputFileTypes,     ApplicationName,
    ApplicationVersion, OptionFiles, DDDB, CONDDB, ExtraPackages,
    Visible, ProcessingPass, Usable, RuntimeProjects, DQTag,
    OptionsFormat, StartItem, MaxItem
    """
    return cls.bkkDB.getAvailableSteps(in_dict)

  #############################################################################
  types_getRuntimeProjects = [dict]

  @classmethod
  def export_getRuntimeProjects(cls, in_dict):
    """It returns a runtime project for a given step.

    The input parameter is a in_dictionary which has only one key StepId
    """
    return cls.bkkDB.getRuntimeProjects(in_dict)

  #############################################################################
  types_getStepInputFiles = [int]

  @classmethod
  def export_getStepInputFiles(cls, stepId):
    """It returns the input files for a given step."""
    retVal = cls.bkkDB.getStepInputFiles(stepId)
    if not retVal['OK']:
      return retVal

    records = [list(record) for record in retVal['Value']]
    return S_OK({'ParameterNames': ['FileType', 'Visible'],
                 'Records': records,
                 'TotalRecords': len(records)})

  #############################################################################
  types_setStepInputFiles = [int, list]

  @classmethod
  def export_setStepInputFiles(cls, stepid, files):
    """It is used to set input file types to a Step."""
    return cls.bkkDB.setStepInputFiles(stepid, files)

  #############################################################################
  types_setStepOutputFiles = [int, list]

  @classmethod
  def export_setStepOutputFiles(cls, stepid, files):
    """It is used to set output file types to a Step."""
    return cls.bkkDB.setStepOutputFiles(stepid, files)

  #############################################################################
  types_getStepOutputFiles = [int]

  @classmethod
  def export_getStepOutputFiles(cls, stepId):
    """It returns the output file types for a given Step."""
    retVal = cls.bkkDB.getStepOutputFiles(stepId)
    if not retVal['OK']:
      return retVal

    records = []
    parameters = ['FileType', 'Visible']
    for record in retVal['Value']:
      records += [list(record)]
    return S_OK({'ParameterNames': parameters, 'Records': records, 'TotalRecords': len(records)})

  #############################################################################
  types_getAvailableFileTypes = []

  @classmethod
  def export_getAvailableFileTypes(cls, ):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getAvailableFileTypes()

  #############################################################################
  types_insertFileTypes = [six.string_types, six.string_types, six.string_types]

  @classmethod
  def export_insertFileTypes(cls, ftype, desc, fileType):
    """It is used to register a file type. It has the following input
    parameters:

    :param str ftype: file type; for example: COOL.DST
    :param str desc: a short description which describes the file content
    :paran str fileType: the file format such as ROOT, POOL_ROOT, etc.
    """
    return cls.bkkDB.insertFileTypes(ftype, desc, fileType)

  #############################################################################
  types_insertStep = [dict]

  @classmethod
  def export_insertStep(cls, in_dict):
    """It used to insert a step to the Bookkeeping Metadata Catalogue. The
    imput parameter is a dictionary which contains the steps attributes. For
    example: Dictionary format:

    {'Step': {'ApplicationName': 'DaVinci',
    'Usable': 'Yes',
    'StepId': '',
    'ApplicationVersion': 'v29r1', 'ext-comp-1273':
    'CHARM.MDST (Charm micro dst)', 'ExtraPackages': '', 'StepName': 'davinci prb2',
    'ProcessingPass': 'WG-Coool', 'ext-comp-1264': 'CHARM.DST (Charm stream)', 'Visible': 'Y', 'DDDB': '',
    'OptionFiles': '', 'CONDDB': ''}, 'OutputFileTypes': [{'Visible': 'Y', 'FileType': 'CHARM.MDST'}],
    'InputFileTypes': [{'Visible': 'Y', 'FileType': 'CHARM.DST'}],'RuntimeProjects':[{StepId:13878}]}
    """
    return cls.bkkDB.insertStep(in_dict)

  #############################################################################
  types_deleteStep = [int]

  @classmethod
  def export_deleteStep(cls, stepid):
    """It used to delete a given step."""
    return cls.bkkDB.deleteStep(stepid)

  #############################################################################
  types_deleteStepContainer = [int]

  @classmethod
  def export_deleteStepContainer(cls, stepid):
    """It used to delete a given step."""
    return cls.bkkDB.deleteStepContainer(stepid)

  #############################################################################
  types_updateStep = [dict]

  @classmethod
  def export_updateStep(cls, in_dict):
    """It is used to modify the step attributes."""
    return cls.bkkDB.updateStep(in_dict)

  ##############################################################################
  types_getAvailableConfigNames = []

  @classmethod
  def export_getAvailableConfigNames(cls):
    """It returns all the available configuration names which are used."""
    retVal = cls.bkkDB.getAvailableConfigNames()
    if not retVal['OK']:
      return retVal

    records = [list(record) for record in retVal['Value']]
    return S_OK({'ParameterNames': ['Configuration Name'],
                 'Records': records,
                 'TotalRecords': len(records)})

  #############################################################################
  types_getConfigVersions = [dict]

  @classmethod
  def export_getConfigVersions(cls, in_dict):
    """It returns all the available configuration version for a given
    condition.

    Input parameter is a dictionary which has the following key: 'ConfigName'
    For example: in_dict = {'ConfigName':'MC'}
    """
    configName = in_dict.get('ConfigName', default)
    retVal = cls.bkkDB.getConfigVersions(configName)
    if not retVal['OK']:
      return retVal
    records = []
    parameters = ['Configuration Version']
    for record in retVal['Value']:
      records += [list(record)]
    return S_OK({'ParameterNames': parameters, 'Records': records, 'TotalRecords': len(records)})

  #############################################################################
  types_getConditions = [dict]

  @classmethod
  def export_getConditions(cls, in_dict):
    """It returns all the available conditions for a given conditions.

    Input parameter is a dictionary which has the following keys: 'ConfigName', 'ConfigVersion', 'EventType'
    For example: in_dict = {'ConfigName':'MC','ConfigVersion':'MC10'}
    """
    result = S_ERROR()
    ok = True
    configName = in_dict.get('ConfigName', default)
    configVersion = in_dict.get('ConfigVersion', default)
    evt = in_dict.get('EventType', in_dict.get('EventTypeId', default))

    if 'EventTypeId' in in_dict:
      gLogger.verbose('EventTypeId will be not accepted! Please change it to EventType')

    retVal = cls.bkkDB.getConditions(configName, configVersion, evt)
    if retVal['OK']:
      values = retVal['Value']
      sim_parameters = ['SimId',
                        'Description',
                        'BeamCondition',
                        'BeamEnergy',
                        'Generator',
                        'MagneticField',
                        'DetectorCondition',
                        'Luminosity',
                        'G4settings']
      daq_parameters = ['DaqperiodId', 'Description',
                        'BeamCondition', 'BeanEnergy',
                        'MagneticField', 'VELO',
                        'IT', 'TT', 'OT', 'RICH1',
                        'RICH2', 'SPD_PRS', 'ECAL',
                        'HCAL', 'MUON', 'L0', 'HLT',
                        'VeloPosition']
      sim_records = []
      daq_records = []

      if len(values) > 0:
        for record in values:
          if record[0] is not None:
            sim_records += [[record[0], record[2],
                             record[3], record[4],
                             record[5], record[6],
                             record[7], record[8],
                             record[9]]]
          elif record[1] is not None:
            daq_records += [[record[1], record[10], record[11],
                             record[12], record[13], record[14],
                             record[15], record[16], record[17],
                             record[18], record[19], record[20],
                             record[21], record[22], record[23],
                             record[24], record[25], record[26]]]
          else:
            result = S_ERROR("Condition does not existis!")
            ok = False
      if ok:
        result = S_OK([{'ParameterNames': sim_parameters,
                        'Records': sim_records,
                        'TotalRecords': len(sim_records)}, {'ParameterNames': daq_parameters,
                                                            'Records': daq_records,
                                                            'TotalRecords': len(daq_records)}
                       ])
    else:
      result = retVal

    return result

  #############################################################################
  types_getProcessingPass = [dict, six.string_types]

  @classmethod
  def export_getProcessingPass(cls, in_dict, path=None):
    """It returns the processing pass for a given conditions.

    Input parameter is a dictionary and a path (string) which has the following keys:
    'ConfigName', 'ConfigVersion', 'ConditionDescription','Production', 'RunNumber', 'EventType'
    This method is used to recursively browse the processing pass.
    To start the browsing you have to define the path as a root: path = '/'
    Note: it returns a list with two dictionary. First dictionary contains the processing passes
    while the second dictionary contains the event types.
    """
    if path is None:
      path = '/'
    configName = in_dict.get('ConfigName', default)
    configVersion = in_dict.get('ConfigVersion', default)
    conddescription = in_dict.get('ConditionDescription', default)
    prod = in_dict.get('Production', default)
    runnb = in_dict.get('RunNumber', default)
    evt = in_dict.get('EventType', in_dict.get('EventTypeId', default))
    return cls.bkkDB.getProcessingPass(configName, configVersion, conddescription, runnb, prod, evt, path)

  #############################################################################
  types_getProductions = [dict]

  @classmethod
  def export_getProductions(cls, in_dict):
    """It returns the productions for a given conditions.

    Input parameter is a dictionary which has the following keys:
    'ConfigName', 'ConfigVersion', 'ConditionDescription',
    'EventType','ProcessingPass'
    """
    configName = in_dict.get('ConfigName', default)
    configVersion = in_dict.get('ConfigVersion', default)
    conddescription = in_dict.get('ConditionDescription', default)
    processing = in_dict.get('ProcessingPass', default)
    evt = in_dict.get('EventType', in_dict.get('EventTypeId', default))
    replicaFlag = in_dict.get('ReplicaFlag', 'Yes')
    ftype = in_dict.get('FileType', default)
    visible = in_dict.get('Visible', 'Y')
    if 'EventTypeId' in in_dict:
      gLogger.verbose('The EventTypeId has to be replaced by EventType!')

    retVal = cls.bkkDB.getProductions(
        configName, configVersion, conddescription, processing, evt, visible, ftype, replicaFlag)
    if not retVal['OK']:
      return retVal
    records = []
    parameters = ['Production/RunNumber']
    for record in retVal['Value']:
      records += [[record[0]]]
    return S_OK({'ParameterNames': parameters, 'Records': records, 'TotalRecords': len(records)})

  #############################################################################
  types_getFileTypes = [dict]

  @classmethod
  def export_getFileTypes(cls, in_dict):
    """It returns the file types for a given conditions.

    Input parameter is a dictionary which has the following keys:
    'ConfigName', 'ConfigVersion', 'ConditionDescription', 'EventType','ProcessingPass','Production','RunNumber'
    """
    configName = in_dict.get('ConfigName', default)
    configVersion = in_dict.get('ConfigVersion', default)
    conddescription = in_dict.get('ConditionDescription', default)
    processing = in_dict.get('ProcessingPass', default)
    evt = in_dict.get('EventType', in_dict.get('EventTypeId', default))
    production = in_dict.get('Production', default)
    runnb = in_dict.get('RunNumber', default)
    visible = in_dict.get('Visible', 'Y')
    replicaflag = in_dict.get('ReplicaFlag', 'Yes')

    if 'EventTypeId' in in_dict:
      gLogger.verbose('The EventTypeId has to be replaced by EventType!')

    retVal = cls.bkkDB.getFileTypes(configName, configVersion,
                                    conddescription, processing,
                                    evt, runnb,
                                    production, visible, replicaflag)
    if not retVal['OK']:
      return retVal
    records = []
    parameters = ['FileTypes']
    for record in retVal['Value']:
      records += [[record[0]]]
    return S_OK({'ParameterNames': parameters, 'Records': records, 'TotalRecords': len(records)})

  #############################################################################
  def transfer_toClient(self, parameters, token, fileHelper):
    """This method used to transfer data using a file.

    Currently two client methods are using this function: getFiles,
    getFilesWithMetadata
    """
    result = S_OK()
    in_dict = JEncoder.loads(parameters)
    gLogger.verbose("The following dictionary received:", "%s" % in_dict)
    methodName = in_dict.get('MethodName', default)
    if methodName == 'getFiles':
      retVal = self.__getFiles(in_dict)
    else:
      retVal = self.__getFilesWithMetadata(in_dict)

    fileString = JEncoder.dumps(retVal)

    retVal = fileHelper.stringToNetwork(fileString)
    if retVal['OK']:
      gLogger.debug('Sent files for', '%s of size %d' % (in_dict, len(fileString)))
    else:
      gLogger.error("Failed to send files:", "%s" % in_dict)
      result = retVal
    return result

  #############################################################################
  @classmethod
  @checkEnoughBKArguments
  def __getFiles(cls, in_dict):
    """It returns a list of files."""
    simdesc = in_dict.get('SimulationConditions', default)
    datataking = in_dict.get('DataTakingConditions', default)
    procPass = in_dict.get('ProcessingPass', default)
    ftype = in_dict.get('FileType', default)
    evt = in_dict.get('EventType', default)
    configname = in_dict.get('ConfigName', default)
    configversion = in_dict.get('ConfigVersion', default)
    prod = in_dict.get('Production', in_dict.get('ProductionID', default))
    flag = in_dict.get('DataQuality', in_dict.get('DataQualityFlag', default))
    startd = in_dict.get('StartDate', None)
    endd = in_dict.get('EndDate', None)
    nbofevents = in_dict.get('NbOfEvents', False)
    startRunID = in_dict.get('StartRun', None)
    endRunID = in_dict.get('EndRun', None)
    runNbs = in_dict.get('RunNumber', in_dict.get('RunNumbers', []))
    if not isinstance(runNbs, list):
      runNbs = [runNbs]
    replicaFlag = in_dict.get('ReplicaFlag', 'Yes')
    visible = in_dict.get('Visible', default)
    filesize = in_dict.get('FileSize', False)
    tck = in_dict.get('TCK', [])
    jobStart = in_dict.get('JobStartDate', None)
    jobEnd = in_dict.get('JobEndDate', None)

    if 'ProductionID' in in_dict:
      gLogger.verbose('ProductionID will be removed. It will changed to Production')

    if 'DataQualityFlag' in in_dict:
      gLogger.verbose('DataQualityFlag will be removed. It will changed to DataQuality')

    if 'RunNumbers' in in_dict:
      gLogger.verbose('RunNumbers will be removed. It will changed to RunNumbers')

    result = []
    retVal = cls.bkkDB.getFiles(simdesc, datataking,
                                procPass, ftype, evt,
                                configname, configversion,
                                prod, flag, startd, endd,
                                nbofevents, startRunID,
                                endRunID, runNbs,
                                replicaFlag, visible,
                                filesize, tck, jobStart, jobEnd)
    if not retVal['OK']:
      result = retVal
    else:
      values = retVal['Value']
      for i in values:
        result += [i[0]]
      result = S_OK(result)

    return result

  #############################################################################
  @classmethod
  @checkEnoughBKArguments
  def __getFilesWithMetadata(cls, in_dict):
    """It returns the files with their metadata."""
    configName = in_dict.get('ConfigName', default)
    configVersion = in_dict.get('ConfigVersion', default)
    conddescription = in_dict.get('ConditionDescription', default)
    processing = in_dict.get('ProcessingPass', default)
    evt = in_dict.get('EventType', in_dict.get('EventTypeId', default))
    production = in_dict.get('Production', default)
    filetype = in_dict.get('FileType', default)
    quality = in_dict.get('DataQuality', in_dict.get('Quality', default))
    visible = in_dict.get('Visible', 'Y')
    replicaFlag = in_dict.get('ReplicaFlag', 'Yes')
    startDate = in_dict.get('StartDate', None)
    endDate = in_dict.get('EndDate', None)
    runnumbers = in_dict.get('RunNumber', in_dict.get('RunNumbers', []))
    startRunID = in_dict.get('StartRun', None)
    endRunID = in_dict.get('EndRun', None)
    tcks = in_dict.get('TCK')
    jobStart = in_dict.get('JobStartDate', None)
    jobEnd = in_dict.get('JobEndDate', None)

    if 'EventTypeId' in in_dict:
      gLogger.verbose('The EventTypeId has to be replaced by EventType!')

    if 'Quality' in in_dict:
      gLogger.verbose('The Quality has to be replaced by DataQuality!')

    retVal = cls.bkkDB.getFilesWithMetadata(configName,
                                            configVersion,
                                            conddescription,
                                            processing,
                                            evt,
                                            production,
                                            filetype,
                                            quality,
                                            visible,
                                            replicaFlag,
                                            startDate,
                                            endDate,
                                            runnumbers,
                                            startRunID,
                                            endRunID,
                                            tcks,
                                            jobStart,
                                            jobEnd)
    if not retVal['OK']:
      return retVal
    records = []
    parameters = ['FileName', 'EventStat', 'FileSize',
                  'CreationDate', 'JobStart', 'JobEnd',
                  'WorkerNode', 'FileType', 'RunNumber',
                  'FillNumber', 'FullStat', 'DataqualityFlag',
                  'EventInputStat', 'TotalLuminosity', 'Luminosity',
                  'InstLuminosity', 'TCK', 'GUID', 'ADLER32', 'EventType', 'MD5SUM',
                  'VisibilityFlag', 'JobId', 'GotReplica', 'InsertTimeStamp']
    for record in retVal['Value']:
      records += [[record[0], record[1], record[2],
                   record[3], record[4], record[5],
                   record[6], record[7], record[8],
                   record[9], record[10], record[11],
                   record[12], record[13], record[14],
                   record[15], record[16], record[17], record[18], record[19],
                   record[20], record[21], record[22], record[23], record[24]]]
    return {'ParameterNames': parameters, 'Records': records, 'TotalRecords': len(records)}

  #############################################################################
  types_getFilesSummary = [dict]

  @checkEnoughBKArguments
  def export_getFilesSummary(self, in_dict):
    """It returns sumary for a given data set.

    Input parameter is a dictionary which has the following keys:
    'ConfigName', 'ConfigVersion', 'ConditionDescription', 'EventType',
    'ProcessingPass','Production','RunNumber', 'FileType', DataQuality
    """
    gLogger.debug('Input:', "%s" % in_dict)

    configName = in_dict.get('ConfigName', default)
    configVersion = in_dict.get('ConfigVersion', default)
    condDescription = in_dict.get('ConditionDescription', default)
    processingPass = in_dict.get('ProcessingPass', default)
    eventType = in_dict.get('EventType', in_dict.get('EventTypeId', default))
    production = in_dict.get('Production', default)
    fileType = in_dict.get('FileType', default)
    dataQuality = in_dict.get('DataQuality', in_dict.get('Quality', default))
    startRun = in_dict.get('StartRun', None)
    endRun = in_dict.get('EndRun', None)
    visible = in_dict.get('Visible', 'Y')
    startDate = in_dict.get('StartDate', None)
    endDate = in_dict.get('EndDate', None)
    runNumbers = in_dict.get('RunNumber', in_dict.get('RunNumbers', []))
    replicaFlag = in_dict.get('ReplicaFlag', 'Yes')
    tcks = in_dict.get('TCK')
    jobStart = in_dict.get('JobStartDate', None)
    jobEnd = in_dict.get('JobEndDate', None)

    if 'EventTypeId' in in_dict:
      gLogger.verbose('The EventTypeId has to be replaced by EventType!')

    if 'Quality' in in_dict:
      gLogger.verbose('The Quality has to be replaced by DataQuality!')

    retVal = self.bkkDB.getFilesSummary(configName=configName,
                                        configVersion=configVersion,
                                        conditionDescription=condDescription,
                                        processingPass=processingPass,
                                        eventType=eventType,
                                        production=production,
                                        fileType=fileType,
                                        dataQuality=dataQuality,
                                        startRun=startRun,
                                        endRun=endRun,
                                        visible=visible,
                                        startDate=startDate,
                                        endDate=endDate,
                                        runNumbers=runNumbers,
                                        replicaFlag=replicaFlag,
                                        tcks=tcks,
                                        jobStart=jobStart,
                                        jobEnd=jobEnd)
    if not retVal['OK']:
      return retVal
    records = []
    parameters = ['NbofFiles', 'NumberOfEvents', 'FileSize', 'Luminosity', 'InstLuminosity']
    for record in retVal['Value']:
      records += [[record[0], record[1], record[2], record[3], record[4]]]
    return S_OK({'ParameterNames': parameters, 'Records': records, 'TotalRecords': len(records)})

  #############################################################################
  types_getLimitedFiles = [dict]

  @classmethod
  def export_getLimitedFiles(cls, in_dict):
    """It returns a chunk of files.

    This method is equivalent to the getFiles.
    """
    configName = in_dict.get('ConfigName', default)
    configVersion = in_dict.get('ConfigVersion', default)
    conddescription = in_dict.get('ConditionDescription', default)
    processing = in_dict.get('ProcessingPass', default)
    evt = in_dict.get('EventType', in_dict.get('EventTypeId', default))
    production = in_dict.get('Production', default)
    filetype = in_dict.get('FileType', default)
    quality = in_dict.get('DataQuality', in_dict.get('Quality', default))
    runnb = in_dict.get('RunNumbers', in_dict.get('RunNumber', default))
    start = in_dict.get('StartItem', 0)
    maxValue = in_dict.get('MaxItem', 10)

    if 'EventTypeId' in in_dict:
      gLogger.verbose('The EventTypeId has to be replaced by EventType!')

    if 'Quality' in in_dict:
      gLogger.verbose('The Quality has to be replaced by DataQuality!')

    retVal = cls.bkkDB.getLimitedFiles(configName,
                                       configVersion,
                                       conddescription,
                                       processing,
                                       evt,
                                       production,
                                       filetype,
                                       quality,
                                       runnb,
                                       start,
                                       maxValue)
    if not retVal['OK']:
      return retVal
    records = []
    parameters = ['Name', 'EventStat', 'FileSize',
                  'CreationDate', 'JobStart', 'JobEnd',
                  'WorkerNode', 'FileType', 'EventType',
                  'RunNumber', 'FillNumber', 'FullStat',
                  'DataqualityFlag', 'EventInputStat',
                  'TotalLuminosity', 'Luminosity',
                  'InstLuminosity', 'TCK', 'WNMJFHS06', 'HLT2TCK',
                  'NumberOfProcessors']
    for record in retVal['Value']:
      records += [[record[0], record[1], record[2],
                   str(record[3]), str(record[4]),
                   str(record[5]), record[6], record[7],
                   record[8], record[9], record[10],
                   record[11], record[12], record[13],
                   record[14], record[15], record[16],
                   record[17]]]
    return S_OK({'ParameterNames': parameters, 'Records': records, 'TotalRecords': len(records)})

  #############################################################################
  types_getAvailableDataQuality = []

  @classmethod
  def export_getAvailableDataQuality(cls):
    """it returns all the available data quality flags."""
    return cls.bkkDB.getAvailableDataQuality()

  #############################################################################
  types_getAvailableProductions = []

  @classmethod
  def export_getAvailableProductions(cls):
    """It returns all the available productions which have associated file with
    replica flag yes."""
    return cls.bkkDB.getAvailableProductions()

  #############################################################################
  types_getAvailableRuns = []

  @classmethod
  def export_getAvailableRuns(cls):
    """It returns all the available runs which have associated files with
    reploica flag yes."""
    return cls.bkkDB.getAvailableRuns()

  #############################################################################
  types_getAvailableEventTypes = []

  @classmethod
  def export_getAvailableEventTypes(cls):
    """It returns all the available event types."""
    return cls.bkkDB.getAvailableEventTypes()

  #############################################################################
  types_getMoreProductionInformations = [int]

  @classmethod
  def export_getMoreProductionInformations(cls, prodid):
    """It returns inforation about a production."""
    return cls.bkkDB.getMoreProductionInformations(prodid)

  #############################################################################
  types_getJobInfo = [six.string_types]

  @classmethod
  def export_getJobInfo(cls, lfn):
    """It returns the job metadata information for a given lfn produced by this
    job."""
    return cls.bkkDB.getJobInfo(lfn)

  #############################################################################
  types_bulkJobInfo = [dict]

  @classmethod
  def export_bulkJobInfo(cls, lfns):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.bulkJobInfo(lfns)

  #############################################################################
  types_getJobInformation = [dict]

  @classmethod
  def export_getJobInformation(cls, in_dict):
    """It returns the job metadata information for a given lfn produced by this
    job."""
    return cls.bkkDB.getJobInformation(in_dict)

  #############################################################################
  types_getRunNumber = [six.string_types]

  @classmethod
  def export_getRunNumber(cls, lfn):
    """It returns the run number for a given lfn!"""
    return cls.bkkDB.getRunNumber(lfn)

  #############################################################################
  types_getRunNbAndTck = [six.string_types]

  @classmethod
  def export_getRunNbAndTck(cls, lfn):
    """It returns the run number and tck for a given LFN."""
    return cls.bkkDB.getRunNbAndTck(lfn)

  #############################################################################
  types_getProductionFiles = [six.integer_types, six.string_types]

  @classmethod
  def export_getProductionFiles(cls, prod, fileType, replica=default):
    """It returns files and their metadata for a given production, file type
    and replica."""
    return cls.bkkDB.getProductionFiles(prod, fileType, replica)

  #############################################################################
  types_getAvailableRunNumbers = []

  def export_getAvailableRunNumbers(self):
    """more info in the BookkeepingClient.py."""
    return self.export_getAvailableRuns()

  #############################################################################
  types_getRunFiles = [int]

  @classmethod
  def export_getRunFiles(cls, runid):
    """It returns all the files and their metadata for a given run number!"""
    return cls.bkkDB.getRunFiles(runid)

  #############################################################################
  types_updateFileMetaData = [six.string_types, dict]

  @classmethod
  def export_updateFileMetaData(cls, filename, fileAttr):
    """This method used to modify files metadata.

    Input parametes is a stirng (filename) and a dictionary (fileAttr)
    with the file attributes. {'GUID':34826386286382,'EventStat':222222}
    """
    return cls.bkkDB.updateFileMetaData(filename, fileAttr)

  #############################################################################
  types_renameFile = [six.string_types, six.string_types]

  @classmethod
  def export_renameFile(cls, oldLFN, newLFN):
    """It allows to change the name of a file which is in the Bookkeeping
    Metadata Catalogue."""
    return cls.bkkDB.renameFile(oldLFN, newLFN)

  #############################################################################
  types_getProductionProcessingPassID = [six.integer_types]

  @classmethod
  def export_getProductionProcessingPassID(cls, prodid):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getProductionProcessingPassID(prodid)

  #############################################################################
  types_getProductionProcessingPass = [six.integer_types]

  @classmethod
  def export_getProductionProcessingPass(cls, prodid):
    """It returns the processing pass for a given production."""
    return cls.bkkDB.getProductionProcessingPass(prodid)

  #############################################################################
  types_insertTag = [dict]

  @classmethod
  def export_insertTag(cls, values):
    """It used to register tags (CONDB, DDDB, etc) to the database.

    The input parameter is dictionary: {'TagName':'Value'}
    """
    successfull = {}
    faild = {}

    for i in values:
      tags = values[i]
      for tag in tags:
        retVal = cls.bkkDB.existsTag(i, tag)
        if retVal['OK'] and not retVal['Value']:
          retVal = cls.bkkDB.insertTag(i, tag)
          if not retVal['OK']:
            faild[tag] = i
          else:
            successfull[tag] = i
        else:
          faild[tag] = i
    return S_OK({'Successfull': successfull, 'Faild': faild})

  #############################################################################
  types_setFileDataQuality = [list, six.string_types]

  @classmethod
  def export_setFileDataQuality(cls, lfns, flag):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.setFileDataQuality(lfns, flag)

  #############################################################################
  types_setRunAndProcessingPassDataQuality = [six.integer_types, six.string_types, six.string_types]

  @classmethod
  def export_setRunAndProcessingPassDataQuality(cls, runNB, procpass, flag):
    """It sets the data quality to a run which belong to a given processing
    pass.

    This method insert a new row to the runquality table. This value
    used to set the data quality flag to a given run files which
    processed by a given processing pass.
    """
    return cls.bkkDB.setRunAndProcessingPassDataQuality(runNB, procpass, flag)

  #############################################################################
  types_setRunDataQuality = [int, six.string_types]

  @classmethod
  def export_setRunDataQuality(cls, runNb, flag):
    """It sets the data quality for a given run!

    The input parameter is the run number and a data quality flag.
    """
    return cls.bkkDB.setRunDataQuality(runNb, flag)

  #############################################################################
  types_setQualityRun = [int, six.string_types]

  @classmethod
  def export_setQualityRun(cls, runNb, flag):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.setRunDataQuality(runNb, flag)

  #############################################################################
  types_setProductionDataQuality = [int, six.string_types]

  @classmethod
  def export_setProductionDataQuality(cls, prod, flag):
    """It sets the data quality for a given production!"""
    return cls.bkkDB.setProductionDataQuality(prod, flag)

  #############################################################################
  types_getFileAncestors = [list, int, bool]

  @classmethod
  def export_getFileAncestors(cls, lfns, depth=None, replica=None):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getFileAncestors(lfns, depth, replica)

  #############################################################################
  types_getAllAncestors = [list, int]

  @classmethod
  def export_getAllAncestors(cls, lfns, depth):
    """more info in the BookkeepingClient.py."""
    retVal = cls.bkkDB.getFileAncestors(lfns, depth, False)
    if not retVal['OK']:
      return retVal
    values = retVal['Value']
    for key, value in values['Successful'].items():
      values['Successful'][key] = [i['FileName'] for i in value]
    return S_OK(values)

  #############################################################################
  types_getAncestors = [list, int]

  @classmethod
  def export_getAncestors(cls, lfns, depth):
    """ Get the ancestors for a list of LFNs in input
    """
    retVal = cls.bkkDB.getFileAncestors(lfns, depth, True)
    if not retVal['OK']:
      return retVal
    values = retVal['Value']
    for key, value in values['Successful'].items():
      values['Successful'][key] = [i['FileName'] for i in value]
    return S_OK(values)

  #############################################################################
  types_getAllAncestorsWithFileMetaData = [list, int]

  @classmethod
  def export_getAllAncestorsWithFileMetaData(cls, lfns, depth):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getFileAncestors(lfns, depth, False)

  #############################################################################
  types_getAllDescendents = [list, int, int, bool]

  @classmethod
  def export_getAllDescendents(cls, lfn, depth=0, production=0, checkreplica=False):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getFileDescendents(lfn, depth, production, checkreplica)

  #############################################################################
  types_getDescendents = [list, int]

  def export_getDescendents(self, lfn, depth):
    """more info in the BookkeepingClient.py."""
    return self.export_getFileDescendants(lfn, depth)

  #############################################################################
  types_getFileDescendents = [list, int, int, bool]

  @classmethod
  def export_getFileDescendents(cls, lfn, depth, production=0, checkreplica=True):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getFileDescendents(lfn, depth, production, checkreplica)

  #############################################################################
  types_getFileDescendants = [list, int, int, bool]

  def export_getFileDescendants(self, lfn, depth, production=0, checkreplica=True):
    """more info in the BookkeepingClient.py."""
    return self.export_getFileDescendents(lfn, depth, production, checkreplica)

  #############################################################################
  types_checkfile = [six.string_types]

  @classmethod
  def export_checkfile(cls, fileName):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.checkfile(fileName)

  #############################################################################
  types_checkFileTypeAndVersion = [six.string_types, six.string_types]

  @classmethod
  def export_checkFileTypeAndVersion(cls, ftype, version):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.checkFileTypeAndVersion(ftype, version)

  #############################################################################
  types_checkEventType = [six.integer_types]

  @classmethod
  def export_checkEventType(cls, eventTypeId):
    """more info in the BookkeepingClient.py."""
    if eventTypeId not in cls.__eventTypeCache:
      retVal = cls.bkkDB.checkEventType(eventTypeId)
      if not retVal['OK']:
        return retVal
      cls.__eventTypeCache[eventTypeId] = retVal

    return cls.__eventTypeCache[eventTypeId]

  #############################################################################
  types_insertSimConditions = [dict]

  @classmethod
  def export_insertSimConditions(cls, in_dict):
    """It inserts a simulation condition to the Bookkeeping Metadata
    catalogue."""
    return cls.bkkDB.insertSimConditions(in_dict)

  #############################################################################
  types_getSimConditions = []

  @classmethod
  def export_getSimConditions(cls):
    """It returns all the simulation conditions which are in the Bookkeeping
    Metadata catalogue."""
    return cls.bkkDB.getSimConditions()

  #############################################################################
  types_removeReplica = [six.string_types]

  @classmethod
  def export_removeReplica(cls, fileName):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.removeReplica(fileName)

  #############################################################################
  types_getFileMetadata = [list]

  @classmethod
  def export_getFileMetadata(cls, lfns):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getFileMetadata(lfns)

  #############################################################################
  types_getFilesInformations = [list]

  @classmethod
  def export_getFilesInformations(cls, lfns):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getFileMetadata(lfns)

  #############################################################################
  types_getFileMetaDataForUsers = [list]

  @classmethod
  def export_getFileMetaDataForUsers(cls, lfns):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getFileMetaDataForWeb(lfns)

  #############################################################################
  types_getFileMetaDataForWeb = [list]

  @classmethod
  def export_getFileMetaDataForWeb(cls, lfns):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getFileMetaDataForWeb(lfns)

  #############################################################################
  types_getProductionFilesForUsers = [int, dict, dict, six.integer_types, six.integer_types]

  @classmethod
  def export_getProductionFilesForUsers(cls, prod, ftype, sortDict, startItem, maxitems):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getProductionFilesForWeb(prod, ftype, sortDict, startItem, maxitems)

  #############################################################################
  types_getProductionFilesForWeb = [six.integer_types, dict, dict, six.integer_types, six.integer_types]

  @classmethod
  def export_getProductionFilesWeb(cls, prod, ftype, sortDict, startItem, maxitems):
    """It returns files and their metadata information for a given
    production."""
    return cls.bkkDB.getProductionFilesForWeb(prod, ftype, sortDict, startItem, maxitems)

  #############################################################################
  types_exists = [list]

  @classmethod
  def export_exists(cls, lfns):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.exists(lfns)

  #############################################################################
  types_addReplica = [list]

  @classmethod
  def export_addReplica(cls, fileName):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.addReplica(fileName)

  #############################################################################
  types_getRunInformations = [six.integer_types]

  @classmethod
  def export_getRunInformations(cls, runnb):
    """It returns run information and statistics."""
    return cls.bkkDB.getRunInformations(runnb)

  #############################################################################
  types_getRunInformation = [dict]

  @classmethod
  def export_getRunInformation(cls, runnb):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getRunInformation(runnb)

  #############################################################################
  types_getFileCreationLog = [six.string_types]

  @classmethod
  def export_getFileCreationLog(cls, lfn):
    """For a given file returns the log files of the job which created it."""
    return cls.bkkDB.getFileCreationLog(lfn)

  #############################################################################
  types_getLogfile = [six.string_types]

  @classmethod
  def export_getLogfile(cls, lfn):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getFileCreationLog(lfn)

  #############################################################################
  types_insertEventType = [six.integer_types, six.string_types, six.string_types]

  @classmethod
  def export_insertEventType(cls, evid, desc, primary):
    """It inserts an event type to the Bookkeeping Metadata catalogue."""
    retVal = cls.bkkDB.checkEventType(evid)
    if not retVal['OK']:  # meaning the event type is not already inserted
      retVal = cls.bkkDB.insertEventTypes(evid, desc, primary)
      if not retVal['OK']:
        return retVal
      return S_OK(str(evid) + ' event type added successfully!')
    return S_OK(str(evid) + ' event type exists')

  #############################################################################
  types_addEventType = [six.integer_types, six.string_types, six.string_types]

  def export_addEventType(self, evid, desc, primary):
    """more info in the BookkeepingClient.py."""
    return self.export_insertEventType(evid, desc, primary)

  #############################################################################
  types_updateEventType = [six.integer_types, six.string_types, six.string_types]

  @classmethod
  def export_updateEventType(cls, evid, desc, primary):
    """It can used to modify an existing event type."""

    retVal = cls.bkkDB.checkEventType(evid)
    if not retVal['OK']:
      return S_ERROR(str(evid) + ' event type is missing in the BKK database!')

    retVal = cls.bkkDB.updateEventType(evid, desc, primary)
    if not retVal['OK']:
      return retVal
    return S_OK(str(evid) + ' event type updated successfully!')

  #############################################################################
  types_addFiles = [list]

  @classmethod
  def export_addFiles(cls, lfns):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.addReplica(lfns)

  #############################################################################
  types_removeFiles = [list]

  @classmethod
  def export_removeFiles(cls, lfns):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.removeReplica(lfns)

  #############################################################################
  types_getProductionSummary = [dict]

  @classmethod
  def export_getProductionSummary(cls, in_dict):
    """It can used to count the number of events for a given dataset."""

    cName = in_dict.get('ConfigName', default)
    cVersion = in_dict.get('ConfigVersion', default)
    production = in_dict.get('Production', default)
    simdesc = in_dict.get('ConditionDescription', default)
    pgroup = in_dict.get('ProcessingPass', default)
    ftype = in_dict.get('FileType', default)
    evttype = in_dict.get('EventType', default)
    return cls.bkkDB.getProductionSummary(cName, cVersion, simdesc, pgroup, production, ftype, evttype)

  #############################################################################
  types_getProductionInformation = [six.integer_types]

  def export_getProductionInformation(self, prodid):
    """It returns statistics (data processing phases, number of events, etc.) for a given production
    """

    nbjobs = None
    nbOfFiles = None
    nbOfEvents = None
    prodinfos = None

    value = self.bkkDB.getProductionNbOfJobs(prodid)
    if value['OK']:
      nbjobs = value['Value']

    value = self.bkkDB.getProductionNbOfFiles(prodid)
    if value['OK']:
      nbOfFiles = value['Value']

    value = self.bkkDB.getProductionNbOfEvents(prodid)
    if value['OK']:
      nbOfEvents = value['Value']

    value = self.bkkDB.getConfigsAndEvtType(prodid)
    if value['OK']:
      prodinfos = value['Value']

    path = '/'

    if not prodinfos:
      self.log.error("No Configs/Event type for production", prodid)
      return S_ERROR("No Configs/Event type")

    cname = prodinfos[0][0]
    cversion = prodinfos[0][1]
    path += cname + '/' + cversion + '/'

    res = self.bkkDB.getProductionSimulationCond(prodid)
    if not res['OK']:
      return res
    path += res['Value']

    res = self.bkkDB.getProductionProcessingPass(prodid)
    if not res['OK']:
      return res
    path += res['Value']
    prefix = '\n' + path

    # FIXME: I think this will crash due to iterating over None if cls.bkkDB.getProductionNbOfEvents(prodid) fails.
    # FIXME: I also have no idea what i is. At at glance I thought it was an integer but its being indexed?
    # FIXME: Why only index 0 and 2? The docstring of getProductionNbOfEvents should probably be fixed.
    for i in nbOfEvents:
      path += prefix + '/' + str(i[2]) + '/' + i[0]
    result = {"Production information": prodinfos,
              "Number of jobs": nbjobs,
              "Number of files": nbOfFiles,
              "Number of events": nbOfEvents,
              'Path': path}
    return S_OK(result)

  #############################################################################
  types_getFileHistory = [six.string_types]

  @classmethod
  def export_getFileHistory(cls, lfn):
    """It returns all the information about a file."""
    retVal = cls.bkkDB.getFileHistory(lfn)
    result = {}
    records = []
    if retVal['OK']:
      values = retVal['Value']
      parameterNames = ['FileId', 'FileName',
                        'ADLER32', 'CreationDate',
                        'EventStat', 'Eventtype',
                        'Gotreplica', 'GUI', 'JobId',
                        'md5sum', 'FileSize', 'FullStat',
                        'Dataquality', 'FileInsertDate',
                        'Luminosity', 'InstLuminosity']
      counter = 0
      for record in values:
        value = [record[0], record[1], record[2],
                 record[3], record[4], record[5],
                 record[6], record[7], record[8],
                 record[9], record[10], record[11],
                 record[12], record[13], record[14],
                 record[15]]
        records += [value]
        counter += 1
      result = {'ParameterNames': parameterNames, 'Records': records, 'TotalRecords': counter}
    else:
      result = S_ERROR(retVal['Message'])
    return S_OK(result)

  #############################################################################
  types_getJobsNb = [six.integer_types]

  @classmethod
  @deprecated("Use getProductionNbOfJobs")
  def export_getJobsNb(cls, prodid):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getProductionNbOfJobs(prodid)

  #############################################################################
  types_getProductionNbOfJobs = [six.integer_types]

  @classmethod
  def export_getProductionNbOfJobs(cls, prodid):
    """It returns the number of jobs for a given production."""
    return cls.bkkDB.getProductionNbOfJobs(prodid)

  #############################################################################
  types_getNumberOfEvents = [six.integer_types]

  @classmethod
  @deprecated("Use getProductionNbOfEvents")
  def export_getNumberOfEvents(cls, prodid):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getProductionNbOfEvents(prodid)

  #############################################################################
  types_getProductionNbOfEvents = [six.integer_types]

  @classmethod
  def export_getProductionNbOfEvents(cls, prodid):
    """It returns the number of events for a given production."""
    return cls.bkkDB.getProductionNbOfEvents(prodid)

  #############################################################################
  types_getSizeOfFiles = [six.integer_types]

  @classmethod
  @deprecated("Use getProductionSizeOfFiles")
  def export_getSizeOfFiles(cls, prodid):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getProductionSizeOfFiles(prodid)

  #############################################################################
  types_getProductionSizeOfFiles = [six.integer_types]

  @classmethod
  def export_getProductionSizeOfFiles(cls, prodid):
    """It returns the size of files for a given production."""
    return cls.bkkDB.getProductionSizeOfFiles(prodid)

  #############################################################################
  types_getNbOfFiles = [six.integer_types]

  @classmethod
  @deprecated("Use getProductionNbOfFiles")
  def export_getNbOfFiles(cls, prodid):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getProductionNbOfFiles(prodid)

  #############################################################################
  types_getProductionNbOfFiles = [six.integer_types]

  @classmethod
  def export_getProductionNbOfFiles(cls, prodid):
    """It returns the number of files produced by a given production."""
    return cls.bkkDB.getProductionNbOfFiles(prodid)

  #############################################################################
  types_getNbOfJobsBySites = [six.integer_types]

  @classmethod
  def export_getNbOfJobsBySites(cls, prodid):
    """It returns the number of jobs executed at different sites for a given
    production."""
    return cls.bkkDB.getNbOfJobsBySites(prodid)

  #############################################################################
  types_getAvailableTags = []

  @classmethod
  def export_getAvailableTags(cls):
    """It returns the available database tags."""
    return cls.bkkDB.getAvailableTags()

  #############################################################################
  types_getProcessedEvents = [six.integer_types]

  @classmethod
  @deprecated("Use getProductionProcessedEvents")
  def export_getProcessedEvents(cls, prodid):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getProductionProcessedEvents(prodid)

  #############################################################################
  types_getProductionProcessedEvents = [six.integer_types]

  @classmethod
  def export_getProductionProcessedEvents(cls, prodid):
    """it returns the number of events processed for a given production."""
    gLogger.debug('getProductionProcessedEvents->Production:', '%d ' % prodid)
    return cls.bkkDB.getProductionProcessedEvents(prodid)

  #############################################################################
  types_getRunsForAGivenPeriod = [dict]

  @classmethod
  def export_getRunsForAGivenPeriod(cls, in_dict):
    """It returns the available runs between a period.

    Input parameters:
    AllowOutsideRuns: If it is true, it only returns the runs which finished before EndDate.
    StartDate: the run start period
    EndDate: the run end period
    CheckRunStatus: if it is true, it check the run is processed or not processed.
    """
    return cls.bkkDB.getRunsForAGivenPeriod(in_dict)

  #############################################################################
  types_getProductionsFromView = [dict]

  @classmethod
  @deprecated("Useless?")
  def export_getProductionsFromView(cls, in_dict):
    """It returns the productions from the bookkeeping view for a given
    processing pass and run number.

    Input parameters: RunNumber ProcessingPass
    """
    # FIXME: might be a useless method
    return cls.bkkDB.getProductionsFromView(in_dict)

  #############################################################################
  types_getRunFilesDataQuality = [list]

  @classmethod
  def export_getRunFilesDataQuality(cls, runs):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getRunFilesDataQuality(runs)

  #############################################################################
  types_setFilesInvisible = [list]

  @classmethod
  def export_setFilesInvisible(cls, lfns):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.setFilesInvisible(lfns)

  #############################################################################
  types_setFilesVisible = [list]

  @classmethod
  def export_setFilesVisible(cls, lfns):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.setFilesVisible(lfns)

  #############################################################################
  types_getRunAndProcessingPassDataQuality = [six.integer_types, six.integer_types]

  @classmethod
  def export_getRunAndProcessingPassDataQuality(cls, runnb, processing):
    """It returns the data quality flag for a given run and processing pass."""
    return cls.bkkDB.getRunAndProcessingPassDataQuality(runnb, processing)

  #############################################################################
  types_getAvailableConfigurations = []

  @classmethod
  def export_getAvailableConfigurations(cls):
    """It returns the available configurations."""
    return cls.bkkDB.getAvailableConfigurations()

  #############################################################################
  types_getRunProcessingPass = [six.integer_types]

  @classmethod
  def export_getRunProcessingPass(cls, runnumber):
    """it returns the run number for a given run."""
    return cls.bkkDB.getRunProcessingPass(runnumber)

  #############################################################################
  types_getProductionFilesStatus = [int, list]

  @classmethod
  def export_getProductionFilesStatus(cls, productionid=None, lfns=None):
    """It returns the file status in the bkk for a given production or a list
    of lfns."""
    if not lfns:
      lfns = []
    return cls.bkkDB.getProductionFilesStatus(productionid, lfns)

  #############################################################################
  types_getFiles = [dict]

  @classmethod
  def export_getFiles(cls, values):
    """more info in the BookkeepingClient.py."""

    simdesc = values.get('SimulationConditions', default)
    datataking = values.get('DataTakingConditions', default)
    procPass = values.get('ProcessingPass', default)
    ftype = values.get('FileType', default)
    evt = values.get('EventType', 0)
    configname = values.get('ConfigName', default)
    configversion = values.get('ConfigVersion', default)
    prod = values.get('Production', values.get('ProductionID', default))
    flag = values.get('DataQuality', values.get('DataQualityFlag', default))
    startd = values.get('StartDate', None)
    endd = values.get('EndDate', None)
    nbofevents = values.get('NbOfEvents', False)
    startRunID = values.get('StartRun', None)
    endRunID = values.get('EndRun', None)
    runNbs = values.get('RunNumber', values.get('RunNumbers', []))
    if not isinstance(runNbs, list):
      runNbs = [runNbs]
    replicaFlag = values.get('ReplicaFlag', 'Yes')
    visible = values.get('Visible', default)
    filesize = values.get('FileSize', False)
    tck = values.get('TCK')
    jobStart = values.get('JobStartDate', None)
    jobEnd = values.get('JobEndDate', None)

    if 'ProductionID' in values:
      gLogger.verbose('ProductionID will be removed. It will changed to Production')

    if 'DataQualityFlag' in values:
      gLogger.verbose('DataQualityFlag will be removed. It will changed to DataQuality')

    if 'RunNumbers' in values:
      gLogger.verbose('RunNumbers will be removed. It will changed to RunNumbers')

    result = []
    retVal = cls.bkkDB.getFiles(simdesc, datataking,
                                procPass, ftype, evt,
                                configname, configversion,
                                prod, flag, startd, endd,
                                nbofevents, startRunID,
                                endRunID, runNbs,
                                replicaFlag, visible,
                                filesize, tck, jobStart, jobEnd)
    if not retVal['OK']:
      return retVal
    values = retVal['Value']
    for i in values:
      result += [i[0]]

    return S_OK(result)

  #############################################################################
  types_getFilesWithGivenDataSetsForUsers = [dict]

  def export_getFilesWithGivenDataSetsForUsers(self, values):
    """more info in the BookkeepingClient.py."""
    return self.export_getVisibleFilesWithMetadata(values)

  #############################################################################
  types_getVisibleFilesWithMetadata = [dict]

  @classmethod
  def export_getVisibleFilesWithMetadata(cls, in_dict):
    """It returns a list of files with metadata for a given condition."""

    conddescription = in_dict.get('SimulationConditions', in_dict.get('DataTakingConditions', default))
    procPass = in_dict.get('ProcessingPass', default)
    ftype = in_dict.get('FileType', default)
    evt = in_dict.get('EventType', default)
    configname = in_dict.get('ConfigName', default)
    configversion = in_dict.get('ConfigVersion', default)
    prod = in_dict.get('Production', in_dict.get('ProductionID', default))
    dqflag = in_dict.get('DataQuality', in_dict.get('DataQualityFlag', default))
    startd = in_dict.get('StartDate', None)
    endd = in_dict.get('EndDate', None)
    startRunID = in_dict.get('StartRun', None)
    endRunID = in_dict.get('EndRun', None)
    runNbs = in_dict.get('RunNumber', in_dict.get('RunNumbers', []))
    replicaFlag = in_dict.get('ReplicaFlag', 'Yes')
    tck = in_dict.get('TCK', [])
    visible = in_dict.get('Visible', 'Y')
    jobStart = in_dict.get('JobStartDate', None)
    jobEnd = in_dict.get('JobEndDate', None)

    if ftype == default:
      return S_ERROR('FileType is missing!')

    if 'ProductionID' in in_dict:
      gLogger.verbose('ProductionID will be removed. It will changed to Production')

    if 'DataQualityFlag' in in_dict:
      gLogger.verbose('DataQualityFlag will be removed. It will changed to DataQuality')

    if 'RunNumbers' in in_dict:
      gLogger.verbose('RunNumbers will be removed. It will changed to RunNumbers')

    gLogger.debug("getVisibleFilesWithMetadata->", "%s" % in_dict)
    result = {}
    retVal = cls.bkkDB.getFilesWithMetadata(configName=configname,
                                            configVersion=configversion,
                                            conddescription=conddescription,
                                            processing=procPass,
                                            evt=evt,
                                            production=prod,
                                            filetype=ftype,
                                            quality=dqflag,
                                            visible=visible,
                                            replicaflag=replicaFlag,
                                            startDate=startd,
                                            endDate=endd,
                                            runnumbers=runNbs,
                                            startRunID=startRunID,
                                            endRunID=endRunID,
                                            tcks=tck,
                                            jobStart=jobStart,
                                            jobEnd=jobEnd)

    summary = 0

    parameters = ['FileName', 'EventStat', 'FileSize',
                  'CreationDate', 'JobStart', 'JobEnd',
                  'WorkerNode', 'FileType', 'RunNumber',
                  'FillNumber', 'FullStat', 'DataqualityFlag',
                  'EventInputStat', 'TotalLuminosity', 'Luminosity',
                  'InstLuminosity', 'TCK', 'GUID', 'ADLER32', 'EventType', 'MD5SUM',
                  'VisibilityFlag', 'JobId', 'GotReplica', 'InsertTimeStamp']

    if not retVal['OK']:
      return retVal

    values = retVal['Value']
    nbfiles = 0
    nbevents = 0
    evinput = 0
    fsize = 0
    tLumi = 0
    lumi = 0
    ilumi = 0
    for i in values:
      nbfiles = nbfiles + 1
      row = dict(zip(parameters, i))
      if row['EventStat'] is not None:
        nbevents += row['EventStat']
      if row['EventInputStat'] is not None:
        evinput += row['EventInputStat']
      if row['FileSize'] is not None:
        fsize += row['FileSize']
      if row['TotalLuminosity'] is not None:
        tLumi += row['TotalLuminosity']
      if row['Luminosity'] is not None:
        lumi += row['Luminosity']
      if row['InstLuminosity'] is not None:
        ilumi += row['InstLuminosity']
      result[row['FileName']] = {'EventStat': row['EventStat'],
                                 'EventInputStat': row['EventInputStat'],
                                 'Runnumber': row['RunNumber'],
                                 'Fillnumber': row['FillNumber'],
                                 'FileSize': row['FileSize'],
                                 'TotalLuminosity': row['TotalLuminosity'],
                                 'Luminosity': row['Luminosity'],
                                 'InstLuminosity': row['InstLuminosity'],
                                 'TCK': row['TCK']}
    if nbfiles > 0:
      summary = {'Number Of Files': nbfiles,
                 'Number of Events': nbevents,
                 'EventInputStat': evinput,
                 'FileSize': fsize / 1000000000.,
                 'TotalLuminosity': tLumi,
                 'Luminosity': lumi,
                 'InstLuminosity': ilumi}
    return S_OK({'LFNs': result, 'Summary': summary})

  #############################################################################
  types_addProduction = [dict]

  @classmethod
  def export_addProduction(cls, infos):
    """It is used to register a production in the bkk.

    Input parameters:
    SimulationConditions
    DataTakingConditions
    Steps: the step which is used to process data for a given production.
    Production:
    InputProductionTotalProcessingPass: it is a path of the input data processing pass
    """

    gLogger.debug("Registering:", infos)
    simcond = infos.get('SimulationConditions', None)
    daqdesc = infos.get('DataTakingConditions', None)
    production = None

    if simcond is None and daqdesc is None:
      return S_ERROR('SimulationConditions and DataTakingConditions are both missing!')

    if 'Steps' not in infos:
      return S_ERROR("Missing Steps!")
    if 'Production' not in infos:
      return S_ERROR('Production is missing!')
    if 'EventType' not in infos:
      return S_ERROR("EventType is missing!")

    steps = infos['Steps']
    inputProdTotalProcessingPass = ''
    production = infos['Production']
    inputProdTotalProcessingPass = infos.get('InputProductionTotalProcessingPass', '')
    configName = infos.get("ConfigName")
    configVersion = infos.get("ConfigVersion")
    eventType = infos.get("EventType")
    return cls.bkkDB.addProduction(production=production,
                                   simcond=simcond,
                                   daq=daqdesc,
                                   steps=steps,
                                   inputproc=inputProdTotalProcessingPass,
                                   configName=configName,
                                   configVersion=configVersion,
                                   eventType=eventType)

  #############################################################################
  types_getEventTypes = [dict]

  @classmethod
  def export_getEventTypes(cls, in_dict):
    """It returns the available event types for a given configuration name and
    configuration version.

    Input parameters: ConfigName, ConfigVersion, Production
    """

    configName = in_dict.get('ConfigName', default)
    configVersion = in_dict.get('ConfigVersion', default)
    production = in_dict.get('Production', default)
    return cls.bkkDB.getEventTypes(configName, configVersion, production)

  #############################################################################
  types_getProcessingPassSteps = [dict]

  @classmethod
  def export_getProcessingPassSteps(cls, in_dict):
    """It returns the steps for a given stepname, processing pass and
    production."""
    stepname = in_dict.get('StepName', default)
    cond = in_dict.get('ConditionDescription', default)
    procpass = in_dict.get('ProcessingPass', default)

    return cls.bkkDB.getProcessingPassSteps(procpass, cond, stepname)

  #############################################################################
  types_getProductionProcessingPassSteps = [dict]

  @classmethod
  def export_getProductionProcessingPassSteps(cls, in_dict):
    """it returns the steps for a given production."""

    if 'Production' in in_dict:
      return cls.bkkDB.getProductionProcessingPassSteps(in_dict['Production'])
    return S_ERROR('The Production dictionary key is missing!!!')

  #############################################################################
  types_getProductionOutputFiles = [dict]

  @deprecated("Use getProductionOutputFileTypes")
  def export_getProductionOutputFiles(self, in_dict):
    """more info in the BookkeepingClient.py."""

    return self.export_getProductionOutputFileTypes(in_dict)

  #############################################################################
  types_getProductionOutputFileTypes = [dict]

  @classmethod
  def export_getProductionOutputFileTypes(cls, in_dict):
    """It returns the output file types which produced by a given
    production."""

    production = in_dict.get('Production', default)
    stepid = in_dict.get('StepId', default)

    if production != default:
      return cls.bkkDB.getProductionOutputFileTypes(production, stepid)
    return S_ERROR('The Production dictionary key is missing!!!')

  #############################################################################
  types_getRunQuality = [six.string_types, six.string_types]

  @deprecated("Use getRunWithProcessingPassAndDataQuality")
  def export_getRunQuality(self, procpass, flag=default):
    """more info in the BookkeepingClient.py."""

    return self.export_getRunWithProcessingPassAndDataQuality(procpass, flag)

  #############################################################################
  types_getRunWithProcessingPassAndDataQuality = [six.string_types, six.string_types]

  @classmethod
  def export_getRunWithProcessingPassAndDataQuality(cls, procpass, flag=default):
    """It returns the run number for a given processing pass and a flag from
    the run quality table."""
    return cls.bkkDB.getRunWithProcessingPassAndDataQuality(procpass, flag)

  #############################################################################
  types_getRuns = [dict]

  @classmethod
  def export_getRuns(cls, in_dict):
    """It returns the runs for a given configuration name and version.

    Input parameters:
    """
    cName = in_dict.get('ConfigName', default)
    cVersion = in_dict.get('ConfigVersion', default)
    if cName != default and cVersion != default:
      return cls.bkkDB.getRuns(cName, cVersion)
    return S_ERROR('The configuration name and version have to be defined!')

  #############################################################################
  types_getRunProcPass = [dict]

  def export_getRunProcPass(self, in_dict):
    """more info in the BookkeepingClient.py."""
    return self.export_getRunAndProcessingPass(in_dict)

  #############################################################################
  types_getRunAndProcessingPass = [dict]

  @classmethod
  def export_getRunAndProcessingPass(cls, in_dict):
    """It returns all the processing pass and run number for a given run."""
    run = in_dict.get('RunNumber', default)
    if run != default:
      return cls.bkkDB.getRunAndProcessingPass(run)
    return S_ERROR('The run number has to be specified!')

  #############################################################################
  types_getProcessingPassId = [six.string_types]

  @classmethod
  def export_getProcessingPassId(cls, fullpath):
    """It returns the ProcessingPassId for a given path.

    this method should not used!
    """
    return cls.bkkDB.getProcessingPassId(fullpath)

  #############################################################################
  types_getRunNbFiles = [dict]

  @deprecated("Use getNbOfRawFiles")
  def export_getRunNbFiles(self, in_dict):
    """more info in the BookkeepingClient.py."""
    return self.export_getNbOfRawFiles(in_dict)

  #############################################################################
  types_getNbOfRawFiles = [dict]

  @classmethod
  def export_getNbOfRawFiles(cls, in_dict):
    """It counts the raw files for a given run and (or) event type."""

    runnb = in_dict.get('RunNumber', default)
    evt = in_dict.get('EventType', in_dict.get('EventTypeId', default))
    if 'EventTypeId' in in_dict:
      gLogger.verbose('The EventTypeId has to be replaced by EventType!')

    replicaFlag = in_dict.get('ReplicaFlag', 'Yes')
    visible = in_dict.get('Visible', 'Y')
    isFinished = in_dict.get("Finished", 'ALL')
    if runnb == default and evt == default:
      return S_ERROR('Run number or event type must be given!')
    retVal = cls.bkkDB.getNbOfRawFiles(runnb, evt, replicaFlag, visible, isFinished)
    if not retVal['OK']:
      return retVal
    return S_OK(retVal['Value'][0][0])

  #############################################################################
  types_getTypeVersion = [list]

  def export_getTypeVersion(self, lfn):
    """more info in the BookkeepingClient.py."""
    return self.export_getFileTypeVersion(lfn)

  #############################################################################
  types_getFileTypeVersion = [list]

  @classmethod
  def export_getFileTypeVersion(cls, lfn):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getFileTypeVersion(lfn)

  #############################################################################
  types_getTCKs = [dict]

  @classmethod
  def export_getTCKs(cls, in_dict):
    """It returns the tcks for a given data set."""
    configName = in_dict.get('ConfigName', default)
    configVersion = in_dict.get('ConfigVersion', default)
    conddescription = in_dict.get('ConditionDescription', default)
    processing = in_dict.get('ProcessingPass', default)
    evt = in_dict.get('EventType', in_dict.get('EventTypeId', default))
    production = in_dict.get('Production', default)
    filetype = in_dict.get('FileType', default)
    quality = in_dict.get('DataQuality', in_dict.get('Quality', default))
    runnb = in_dict.get('RunNumber', default)
    if 'Quality' in in_dict:
      gLogger.verbose('The Quality has to be replaced by DataQuality!')

    if 'EventTypeId' in in_dict:
      gLogger.verbose('The EventTypeId has to be replaced by EventType!')

    retVal = cls.bkkDB.getTCKs(configName,
                               configVersion,
                               conddescription,
                               processing,
                               evt,
                               production,
                               filetype,
                               quality,
                               runnb)
    if not retVal['OK']:
      return retVal
    return S_OK([record[0] for record in retVal['Value']])

  #############################################################################
  types_getAvailableTcks = [dict]

  def export_getAvailableTcks(self, in_dict):
    """more info in the BookkeepingClient.py."""
    return self.export_getTCKs(in_dict)

  #############################################################################
  types_getSteps = [six.integer_types]

  @classmethod
  def export_getSteps(cls, prodID):
    """ get list of steps used in a production
    """
    return cls.bkkDB.getSteps(prodID)

  #############################################################################
  types_getStepsMetadata = [dict]

  @classmethod
  def export_getStepsMetadata(cls, in_dict):
    """It returns the step(s) which is produced  a given dataset."""
    gLogger.debug('getStepsMetadata', "%s" % in_dict)
    configName = in_dict.get('ConfigName', default)
    configVersion = in_dict.get('ConfigVersion', default)
    cond = in_dict.get('ConditionDescription', default)
    procpass = in_dict.get('ProcessingPass', default)
    evt = in_dict.get('EventType', in_dict.get('EventTypeId', default))
    production = in_dict.get('Production', default)
    filetype = in_dict.get('FileType', default)
    runnb = in_dict.get('RunNumber', default)

    if 'EventTypeId' in in_dict:
      gLogger.verbose('The EventTypeId has to be replaced by EventType!')

    if 'Quality' in in_dict:
      gLogger.verbose('The Quality has to be replaced by DataQuality!')

    return cls.bkkDB.getStepsMetadata(configName, configVersion, cond, procpass, evt, production, filetype, runnb)

  #############################################################################
  types_getDirectoryMetadata_new = [list]

  @classmethod
  @deprecated("Use getDirectoryMetadata")
  def export_getDirectoryMetadata_new(cls, lfn):
    """more info in the BookkeepingClient.py."""
    return cls.bkkDB.getDirectoryMetadata(lfn)

    #############################################################################
  types_getDirectoryMetadata = [list]

  @classmethod
  def export_getDirectoryMetadata(cls, lfn):
    """more info in the BookkeepingClient.py."""
    gLogger.verbose("Getting the metadata for:", "%s" % lfn)
    return cls.bkkDB.getDirectoryMetadata(lfn)

  #############################################################################
  types_getFilesForGUID = [six.string_types]

  @classmethod
  def export_getFilesForGUID(cls, guid):
    """It returns a file for a given GUID."""
    return cls.bkkDB.getFilesForGUID(guid)

  #############################################################################
  types_getRunsGroupedByDataTaking = []

  @classmethod
  def export_getRunsGroupedByDataTaking(cls):
    """It returns all the run numbers grouped by the data taking
    description."""
    return cls.bkkDB.getRunsGroupedByDataTaking()

  #############################################################################
  types_getListOfFills = [dict]

  @classmethod
  def export_getListOfFills(cls, in_dict):
    """It returns a list of FILL numbers for a given Configuration name,
    Configuration version and data taking description."""
    configName = in_dict.get('ConfigName', default)
    configVersion = in_dict.get('ConfigVersion', default)
    conddescription = in_dict.get('ConditionDescription', default)
    return cls.bkkDB.getListOfFills(configName, configVersion, conddescription)

  #############################################################################
  types_getRunsForFill = [six.integer_types]

  @classmethod
  def export_getRunsForFill(cls, fillid):
    """It returns a list of runs for a given FILL."""
    return cls.bkkDB.getRunsForFill(fillid)

  #############################################################################
  types_getListOfRuns = [dict]

  @classmethod
  def export_getListOfRuns(cls, in_dict):
    """It returns a list of runs for a given conditions.

    Input parameter is a dictionary which has the following keys:
    'ConfigName', 'ConfigVersion', 'ConditionDescription',
    'EventType','ProcessingPass'
    """
    configName = in_dict.get('ConfigName', default)
    configVersion = in_dict.get('ConfigVersion', default)
    conddescription = in_dict.get('ConditionDescription', default)
    processing = in_dict.get('ProcessingPass', default)
    evt = in_dict.get('EventType', default)
    quality = in_dict.get('DataQuality', default)

    retVal = cls.bkkDB.getListOfRuns(configName, configVersion, conddescription, processing, evt, quality)
    if not retVal['OK']:
      return retVal
    return S_OK([i[0] for i in retVal['Value']])

  #############################################################################
  types_getSimulationConditions = [dict]

  @classmethod
  def export_getSimulationConditions(cls, in_dict):
    """It returns a list of simulation conditions for a given conditions."""
    return cls.bkkDB.getSimulationConditions(in_dict)

  #############################################################################
  types_updateSimulationConditions = [dict]

  @classmethod
  def export_updateSimulationConditions(cls, in_dict):
    """It updates a given simulation condition."""
    return cls.bkkDB.updateSimulationConditions(in_dict)

  #############################################################################
  types_deleteSimulationConditions = [six.integer_types]

  @classmethod
  def export_deleteSimulationConditions(cls, simid):
    """deletes a given simulation conditions."""
    return cls.bkkDB.deleteSimulationConditions(simid)

  #############################################################################
  types_getProductionSummaryFromView = [dict]

  @classmethod
  def export_getProductionSummaryFromView(cls, in_dict):
    """it returns a summary for a given condition."""
    return cls.bkkDB.getProductionSummaryFromView(in_dict)

  types_getJobInputOutputFiles = [list]

  @classmethod
  def export_getJobInputOutputFiles(cls, diracjobids):
    """It returns the input and output files for a given DIRAC jobid."""
    return cls.bkkDB.getJobInputOutputFiles(diracjobids)

  types_setRunOnlineFinished = [six.integer_types]

  @classmethod
  def export_setRunOnlineFinished(cls, runnumber):
    """It is used to set the run finished..."""
    return cls.bkkDB.setRunStatusFinished(runnumber, 'Y')

  types_setRunOnlineNotFinished = [six.integer_types]

  @classmethod
  def export_setRunOnlineNotFinished(cls, runnumber):
    """You can set the runs not finished."""
    return cls.bkkDB.setRunStatusFinished(runnumber, 'N')

  types_getRunStatus = [list]

  @classmethod
  def export_getRunStatus(cls, runnumbers):
    """it returns the status of the runs."""
    return cls.bkkDB.getRunStatus(runnumbers)

  types_bulkupdateFileMetaData = [dict]

  @classmethod
  def export_bulkupdateFileMetaData(cls, lfnswithmeta):
    """It updates the file metadata."""
    return cls.bkkDB.bulkupdateFileMetaData(lfnswithmeta)

  types_fixRunLuminosity = [list]

  @classmethod
  def export_fixRunLuminosity(cls, runnumbers):
    return cls.bkkDB.fixRunLuminosity(runnumbers)

  #############################################################################
  types_getProductionProducedEvents = [six.integer_types]

  @classmethod
  def export_getProductionProducedEvents(cls, prodid):
    """it returns the number of events producced for a given production."""
    gLogger.debug("Retrieving the number of processed event for production", prodid)
    return cls.bkkDB.getProductionProducedEvents(prodid)

  #############################################################################
  types_bulkinsertEventType = [list]

  @classmethod
  def export_bulkinsertEventType(cls, eventtypes):
    """It inserts a list of event types to the db.

    :param eventtypes: it is a list of event types. For example, the list elements are the following

      .. code-block:: python

        {'EVTTYPEID': '12265021',
         'DESCRIPTION': 'Bu_D0pipipi,Kpi-withf2=DecProdCut_pCut1600MeV',
         'PRIMARY': '[B+ -> (D~0 -> K+ pi-) pi+ pi- pi+]cc'}


    :return: S_ERROR S_OK({'Failed':[],'Successful':[]})
    """
    return cls.bkkDB.bulkinsertEventType(eventtypes)

  #############################################################################
  types_bulkupdateEventType = [list]

  @classmethod
  def export_bulkupdateEventType(cls, eventtypes):
    """It updates a list of event types which are exist in the db.

    :param list eventtypes: it is a list of event types. For example: the list elements are the following:

      .. code-block:: python

      {'EVTTYPEID': '12265021',
      'DESCRIPTION': 'Bu_D0pipipi,Kpi-withf2=DecProdCut_pCut1600MeV',
      'PRIMARY': '[B+ -> (D~0 -> K+ pi-) pi+ pi- pi+]cc'}

    :return: S_ERROR S_OK({'Failed':[],'Successful':[]})
    """
    return cls.bkkDB.bulkupdateEventType(eventtypes)

  #############################################################################
  types_getRunConfigurationsAndDataTakingCondition = [six.integer_types]

  @classmethod
  def export_getRunConfigurationsAndDataTakingCondition(cls, runnumber):
    """It returns minimal information for a given run.

    :param: int runnumber
    :return: S_OK()/S_ERROR ConfigName, ConfigVersion and DataTakingDescription
    """
    return cls.bkkDB.getRunConfigurationsAndDataTakingCondition(runnumber)

  types_deleteCertificationData = []

  @classmethod
  def export_deleteCertificationData(cls):
    """It destroy the data used by the integration test."""
    return cls.bkkDB.deleteCertificationData()

  types_updateProductionOutputfiles = []

  @classmethod
  def export_updateProductionOutputfiles(cls):
    """It is used to trigger an update of the productionoutputfiles table."""
    return cls.bkkDB.updateProductionOutputfiles()

  types_getAvailableTagsFromSteps = []

  @classmethod
  def export_getAvailableTagsFromSteps(cls):
    """It returns the all used datatbase tags: DDDB, CondDB, DQTag."""
    return cls.bkkDB.getAvailableTagsFromSteps()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unittest for:
LHCbDIRAC.HCbDIRAC.Core.Utilities.GeneratorLog"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pytest
from LHCbDIRAC.Core.Utilities.GeneratorLog import counterJson, efficiencyJson, fractionJson, crossSectionJson, methodGeneratorJson  # noqa


listCounters = [{'@name': 'all events (including empty events)', 'value': '556'},
                {'@name': 'events with 0 interaction', 'value': '51'},
                {'@name': 'generated events', 'value': '505'},
                {'@name': 'generated interactions', 'value': '1395'},
                {'@name': 'generated interactions with >= 1b', 'value': '7'},
                {'@name': 'generated interactions with >= 3b', 'value': '1'},
                {'@name': 'generated interactions with 1 prompt B', 'value': '0'},
                {'@name': 'generated interactions with >= 1c', 'value': '91'},
                {'@name': 'generated interactions with >= 3c', 'value': '10'},
                {'@name': 'generated interactions with >= prompt C', 'value': '2'},
                {'@name': 'generated interactions with b and c', 'value': '1'},
                {'@name': 'accepted events', 'value': '1'}]

dictCounters = {
    'accepted events': 1,
    'all events (including empty events)': 556,
    'events with 0 interaction': 51,
    'generated events': 505,
    'generated interactions': 1395,
    'generated interactions with 1 prompt B': 0,
    'generated interactions with >= 1b': 7,
    'generated interactions with >= 1c': 91,
    'generated interactions with >= 3b': 1,
    'generated interactions with >= 3c': 10,
    'generated interactions with >= prompt C': 2,
    'generated interactions with b and c': 1
}


@pytest.mark.parametrize("test_input,expected", [(listCounters, dictCounters),
                                                 ([], {})])
def test_counterJson(test_input, expected):
  assert counterJson(test_input) == expected


listEfficiencies = [
    {
        '@name': 'generator level cut',
        'after': '1',
        'before': '4',
        'error': '0.21651',
        'value': '0.25'
    },
    {
        '@name': 'generator particle level cut',
        'after': '0',
        'before': '1',
        'error': '0',
        'value': '0'
    },
    {
        '@name': 'generator anti-particle level cut',
        'after': '1',
        'before': '3',
        'error': '0.27217',
        'value': '0.33333'
    }
]

dictEfficiencies = {
    'generator anti-particle level cut': {
        'after': 1,
        'before': 3,
        'error': 0.27217,
        'value': 0.33333
    },
    'generator level cut': {
        'after': 1,
        'before': 4,
        'error': 0.21651,
        'value': 0.25
    },
    'generator particle level cut': {
        'after': 0,
        'before': 1,
        'error': 0.0,
        'value': 0.0
    }
}


@pytest.mark.parametrize("test_input,expected", [(listEfficiencies, dictEfficiencies),
                                                 ([], {})])
def test_efficiencyJson(test_input, expected):
  assert efficiencyJson(test_input) == expected


listFractions = [{'@name': 'accepted B+', 'error': '0.35355', 'number': '1', 'value': '0.5'},
                 {'@name': 'accepted B-', 'error': '0', 'number': '1', 'value': '1'},
                 {'@name': 'accepted B0', 'error': '0.35355', 'number': '1', 'value': '0.5'}]

dictFractions = {
    'accepted B+': {
        'error': 0.35355,
        'number': 1,
        'value': 0.5
    },
    'accepted B-': {
        'error': 0.0,
        'number': 1,
        'value': 1.0
    },
    'accepted B0': {
        'error': 0.35355,
        'number': 1,
        'value': 0.5
    }
}


@pytest.mark.parametrize("test_input,expected", [(listFractions, dictFractions),
                                                 ([], {})])
def test_fractionJson(test_input, expected):
  assert fractionJson(test_input) == expected


listCrossSections = [
    {
        '@id': '101',
        'description': '"non-diffractive"',
        'generated': '776',
        'value': '52.116'
    },
    {
        '@id': '102',
        'description': '"A B -> A B elastic"',
        'generated': '300',
        'value': '19.892'
    },
    {
        '@id': '103',
        'description': '"A B -> X B single diffractive"',
        'generated': '120',
        'value': '6.24'
    }
]

dictCrossSections = {
    'A B -> A B elastic': {
        'ID': 102,
        'generated': 300,
        'value': 19.892
    },
    'A B -> X B single diffractive': {
        'ID': 103,
        'generated': 120,
        'value': 6.24
    },
    'non-diffractive': {
        'ID': 101,
        'generated': 776,
        'value': 52.116
    }
}


@pytest.mark.parametrize("test_input,expected", [(listCrossSections, dictCrossSections),
                                                 ([], {})])
def test_crossSectionJson(test_input, expected):
  assert crossSectionJson(test_input) == expected


listMethods = ['GenerationSignal.SignalPlain', 'Generation.SignalRepeatedHadronization']
listGenerators = ['Redecay', 'Pythia8']
dictMethodsGenerators = {
    'GenerationSignal.SignalPlain': 'Redecay',
    'Generation.SignalRepeatedHadronization': 'Pythia8'
}

Method = 'Generation.SignalRepeatedHadronization'
Generator = 'Pythia8'
dictMethodGenerator = {'Generation.SignalRepeatedHadronization': 'Pythia8'}


@pytest.mark.parametrize("test_input1,test_input2,test_input3,expected", [(listMethods, listGenerators, 2, dictMethodsGenerators),  # noqa
                                                                          (Method, Generator, 1, dictMethodGenerator)])
def test_methodGeneratorJson(test_input1, test_input2, test_input3, expected):
  assert methodGeneratorJson(test_input1, test_input2, test_input3) == expected



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit tests for LHCbDIRAC RunApplication module."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from mock import MagicMock

from LHCbDIRAC.Core.Utilities.RunApplication import RunApplication


def test_lbRunCommand():
  """Testing lb-run command (for setting the environment)"""
  ra = RunApplication()
  ra.extraPackages = [('package1', 'v1r0'), ('package2', 'v2r0'), ('package3', '')]
  ra.runTimeProject = 'aRunTimeProject'
  ra.runTimeProjectVersion = 'v1r1'
  ra.opsH = MagicMock()
  ra.opsH.getValue.return_value = ['lcg1', 'lcg2']
  ra.prodConf = True
  extraPackagesString, runtimeProjectString, externalsString = ra._lbRunCommandOptions()
  assert extraPackagesString == ' --use="package1 v1r0"  --use="package2 v2r0"  --use="package3"'
  assert runtimeProjectString == ' --runtime-project aRunTimeProject/v1r1'
  assert externalsString == ' --ext=lcg1 --ext=lcg2'

  ra.site = 'Site1'
  extraPackagesString, runtimeProjectString, externalsString = ra._lbRunCommandOptions()
  assert extraPackagesString == ' --use="package1 v1r0"  --use="package2 v2r0"  --use="package3"'
  assert runtimeProjectString == ' --runtime-project aRunTimeProject/v1r1'
  assert externalsString == ' --ext=lcg1 --ext=lcg2'


def test__gaudirunCommand(mocker):
  """Testing what is run (the gaudirun command, for example)"""
  ra = RunApplication()
  ra.opsH = MagicMock()
  ra.opsH.getValue.return_value = 'gaudirun.py'

  # simplest
  res = str(ra._gaudirunCommand())
  expected = 'gaudirun.py'
  assert res == expected

  # simplest with extra opts
  ra.extraOptionsLine = 'bla bla'
  res = str(ra._gaudirunCommand())
  expected = 'gaudirun.py gaudi_extra_options.py'
  assert res == expected
  with open('gaudi_extra_options.py', 'r') as fd:
    geo = fd.read()
    assert geo == ra.extraOptionsLine

  # productions style /1
  ra.prodConf = True
  ra.extraOptionsLine = ''
  ra.prodConfFileName = 'prodConf.py'
  res = str(ra._gaudirunCommand())
  expected = 'gaudirun.py prodConf.py'
  assert res == expected

  # productions style /2 (multicore)
  ra.optFile = ''
  ra.numberOfProcessors = 2
  res = str(ra._gaudirunCommand())
  expected = 'gaudirun.py --ncpus 2  prodConf.py'
  assert res == expected  # it won't be allowed on this "CE"

  # productions style /3 (multicore and opts)
  ra.optFile = ''
  ra.numberOfProcessors = 2
  ra.extraOptionsLine = 'bla bla'
  res = str(ra._gaudirunCommand())
  expected = 'gaudirun.py --ncpus 2  prodConf.py gaudi_extra_options.py'
  assert res == expected  # it won't be allowed on this "CE"

  # productions style /4
  ra.extraOptionsLine = ''
  ra.commandOptions = ['$APP/1.py',
                       '$APP/2.py']
  res = str(ra._gaudirunCommand())
  expected = r'gaudirun.py --ncpus 2  $APP/1.py $APP/2.py prodConf.py'
  assert res == expected



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Restart any agent  and service installed on a VOBOX."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import re
import os
import sys

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  if 'DIRAC' not in os.environ:
    print("The DIRAC environment is not set")
    sys.exit(0)
  else:
    diracroot = os.environ['DIRAC']

  if os.path.isdir(os.path.join(diracroot, 'runit')) and os.path.isdir(os.path.join(diracroot, 'startup')):
    diracrunit = os.path.join(diracroot, 'runit')
    diracstartup = os.path.join(diracroot, 'startup')
    for link in os.listdir(diracstartup):
      system = link.split('_')[0]
      agent = link.split('_')[1]
      if re.search('Agent', link):
        if not os.path.isdir(os.path.join(diracrunit, system)):
          os.mkdir(os.path.join(diracrunit, system))

        diracsystem = os.path.join(diracrunit, system)
        if not os.path.isdir(os.path.join(diracroot, 'control', system, agent)):
          print(diracsystem)
          print(agent)
          os.mkdir(os.path.join(diracroot, 'control', system, agent))

        print('Restart Agent ' + agent)
        filename_stop = os.path.join(diracroot, 'control', system, agent, 'stop_agent')
        print(filename_stop)
        with open(filename_stop, 'w'):
          pass
      else:
        if re.search('Framework_SystemAdministrator', link):
          print('Skip Framework_SystemAdministrator')
        else:
          print('Restart Service ' + os.path.join(diracstartup, link))
          os.system('runsvctrl t ' + os.path.join(diracstartup, link))

  sys.exit(0)


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Create a new dataset replication or removal transformation according to
plugin."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from LHCbDIRAC.TransformationSystem.Utilities.PluginScript import PluginScript, Script

  pluginScript = PluginScript()
  pluginScript.registerPluginSwitches()
  pluginScript.registerFileSwitches()
  Script.registerSwitch("", "Name=", "   Give a name to the transformation, only if files are given")
  Script.registerSwitch(
      "",
      "SetInvisible",
      "Before creating the transformation, set the files in the BKQuery as invisible (default for DeleteDataset)")
  Script.registerSwitch("S", "Start", "   If set, the transformation is set Active and Automatic [False]")
  Script.registerSwitch("", "Force", "   Force transformation to be submitted even if no files found")
  Script.registerSwitch("", "Test", "   Just print out but not submit")
  Script.registerSwitch("", "NoFCCheck", "   Suppress the check in FC for removal transformations")
  Script.registerSwitch("", "Unique", "   Refuses to create a transformation with an existing name")
  Script.registerSwitch("", "Depth=", "   Depth in path for replacing /... in processing pass")
  Script.registerSwitch("", "Chown=", "   Give user/group for chown of the directories of files in the FC")
  Script.registerSwitch(
      "",
      "MCVersion=",
      "   (list of) BK ConfigVersion; gets active MC processing passes ('All' for all years)")
  Script.registerSwitch("", "ListProcessingPasses", "   Only lists the processing passes")

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ...' % Script.scriptName, ]))

  Script.parseCommandLine(ignoreErrors=True)

  from LHCbDIRAC.DataManagementSystem.Client.AddTransformation import executeAddTransformation
  executeAddTransformation(pluginScript)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.ResourceStatusSystem.Agent.ShiftDBAgent.

ShiftDBAgent.__bases__:
  DIRAC.Core.Base.AgentModule.AgentModule
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# FIXME: should add a "DryRun" option to run in certification setup
from six.moves.urllib.request import urlopen
from six.moves.urllib.error import URLError
import json
import suds.client

from DIRAC import gConfig, S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

AGENT_NAME = 'ResourceStatus/ShiftDBAgent'


class ShiftDBAgent(AgentModule):
  """ This agent queries the LHCb ShiftDB and gets the emails of each piquet
      Then, populates the eGroup associated
      The e-groups admin should be : lhcb-grid-experiment-egroup-admins
  """

  def __init__(self, *args, **kwargs):

    AgentModule.__init__(self, *args, **kwargs)

    # Members initialization

    # ShiftDB url where to find shifter emails
    self.lbshiftdburl = 'https://lbshiftdb.cern.ch/list_email'
    # soap wsdl to access eGroups
    self.wsdl = 'https://foundservices.cern.ch/ws/egroups/v1/EgroupsWebService/EgroupsWebService.wsdl'

    self.roles = {}
    self.roleShifters = {}
    self.newShifters = {}

    self.diracAdmin = None

  def initialize(self, *args, **kwargs):
    """Initialize."""

    self.lbshiftdburl = self.am_getOption('lbshiftdburl', self.lbshiftdburl)
    self.wsdl = self.am_getOption('wsdl', self.wsdl)

    self.user = self.am_getOption('user')
    self.passwd = self.am_getOption('password')

    if not (self.user and self.passwd):
      self.log.error("User and/or password for %s not provided" % self.wsdl)
      return S_ERROR("Creds not provided")

    self.diracAdmin = DiracAdmin()

    return S_OK()

  def execute(self):
    """Execution."""

    self.roles = {}
    self.roleShifters = {}
    self.newShifters = {}

    self.log.info('Getting roles from CS')
    roles = self.__getRolesFromCS()
    if not roles['OK']:
      self.log.error(roles['Message'])
      return roles
    self.log.info('found %s ' % ', '.join(roles['Value']))

    self.log.info('Getting role emails')

    for role, eGroup in roles['Value'].items():

      self.roles[role] = eGroup

      email = self.__getRoleEmail(role)
      if not email['OK']:
        self.log.error(email['Message'])
        # We do not return, we keep execution to clean old shifters
        email['Value'] = None

      email = email['Value']
      self.roleShifters[eGroup] = (email, role)

      self.log.info('%s -> %s' % (role, email))

    self.log.info('Setting role emails')
    for eGroup, roleTuple in self.roleShifters.items():

      email, role = roleTuple

      setEmail = self.__setRoleEmail(eGroup, email, role)
      if not setEmail['OK']:
        self.log.error(setEmail['Message'])

    for newShifterRole, shifterEgroup in self.newShifters.items():

      self.log.info('Notifying role %s' % newShifterRole)
      res = self.__notifyNewShifter(newShifterRole, shifterEgroup)
      if not res['OK']:
        self.log.error(res['Message'])

    return S_OK()

  def __getRolesFromCS(self):
    """ Gets from the CS the roles we want to add to an eGroup.
        Role1 : { eGroup: egroup-blah } in the CS

        returns S_OK( { role1: egroup1, .. } )
    """

    _section = self.am_getModuleParam('section')
    roles = gConfig.getSections('%s/roles' % _section)

    if not roles['OK']:
      return roles

    eGroups = {}

    for role in roles['Value']:
      eGroup = gConfig.getValue('%s/roles/%s/eGroup' % (_section, role))
      if eGroup:
        eGroups[role] = eGroup
        self.log.debug('Found %s : %s ' % (role, eGroup))

    return S_OK(eGroups)

  def __getRoleEmail(self, role):
    """Get role email from shiftDB."""

    try:
      web = urlopen(self.lbshiftdburl, timeout=60)
    except URLError as e:
      return S_ERROR('Cannot open URL: %s, erorr %s' % (self.lbshiftdburl, e))

    emailperson = []

    for line in web.readlines():
      for item in json.loads(line):
          if role in item['role']:
            # There are three shifts per day, so we take into account what time is it
            # before sending the email.
            emailperson = item['email']
            self.log.info(emailperson)
            return S_OK(emailperson)

    return S_ERROR('Email not found')

  def __setRoleEmail(self, eGroup, email, role):
    """Set email in eGroup."""

    client = suds.client.Client(self.wsdl, username=self.user, password=self.passwd)

    try:
      wgroup = client.service.FindEgroupByName(eGroup)
    except suds.WebFault as wError:
      return S_ERROR(wError)

    members = []
    lastShifterEmail = []
    lastShifterList = {}

    if hasattr(wgroup, 'warnings'):
      if wgroup.warnings != []:
        self.log.warn(wgroup.warnings)
    #  return S_ERROR(wgroup.warnings)
    if hasattr(wgroup, 'error'):
      if wgroup.error != []:
        self.log.error(wgroup.error)
    if hasattr(wgroup, 'result') and hasattr(wgroup.result, 'Members'):
      for members in wgroup.result.Members:
        lastShifterEmail.append(members.Email)
        lastShifterList[members.Email] = members

    if email is None:
      self.log.warn("None email. Keeping previous one till an update is found.")
      return S_OK()

#    if lastShifterEmail.strip().lower() == email.strip().lower():
#      self.log.info( "%s has not changed as shifter, no changes needed" % email )
#      return S_OK()

    for lastShifter in lastShifterEmail:
      if email.count(lastShifter) == 0:
        self.log.info("%s is not anymore shifter, deleting ..." % lastShifter)
        try:
          # The last boolean flag is to overwrite
          client.service.RemoveEgroupMembers(eGroup, lastShifterList[lastShifter])
        except suds.WebFault as wError:
          return S_ERROR(wError)

    # Adding a member means it will be the only one in the eGroup, as it is overwritten
    if email != []:
      res = self.__addMember(email, client, eGroup)
      self.log.info(email)
      if not res['OK']:
        self.log.error(res['Message'])
        return res
      self.log.info("%s added successfully to the eGroup for role %s" % (email, role))

    self.newShifters[role] = eGroup

    return S_OK()

  def __addMember(self, email, client, wgroup):
    """Adds a new member to the group."""

    # Clear e-Group before inserting anything
    # self.__deleteMembers( client, wgroup )

    self.log.info('Adding member %s to eGroup %s' % (email, wgroup))

    members = []
    newmember = client.factory.create('ns0:MemberType')
    newmember.Type = "External"
    newmember.Email = email

    members.append(newmember)

    try:
      # The last boolean flag is to overwrite
      self.log.info(members)
      client.service.AddEgroupMembers(wgroup, True, members)
    except suds.WebFault as wError:
      return S_ERROR(wError)
    return S_OK()

  def __notifyNewShifter(self, role, eGroup):
    """Sends an email to the shifter ( if any ) at the beginning of the shift
    period."""

    if role == 'Production':
      body = __productionBody__
    else:
      self.log.info('No email body defined for %s role' % role)
      return S_OK()

    prodRole = self.roles['Production']
    geocRole = self.roles['Grid Expert']
    body = body % (self.roleShifters[prodRole][0], self.roleShifters[geocRole][0])

    # Hardcoded Concezio's email to avoid dirac@mail.cern.ch be rejected by smtp server
    res = self.diracAdmin.sendMail('%s@cern.ch' % eGroup, 'Shifter information',
                                   body, fromAddress='concezio.bozzi@cern.ch')
    return res


__productionBody__ = '''Dear GEOC,

this is an (automatic) mail to welcome you on the grid operations shifts.
In order to facilitate your shift activities we wanted to provide you some pointers,
where you could find more information about shifts, the activities and your duties during this period.

http://lhcb-shifters.web.cern.ch/


LHCbDIRAC portal
https://lhcb-portal-dirac.cern.ch/DIRAC


The logbook for LHCb operations, where all activities concerning offline operation are being logged.
https://lblogbook.cern.ch/Operations/
'''



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve an access URL for an LFN replica given a valid DIRAC SE."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, Script

  dmScript = DMScript()
  dmScript.registerFileSwitches()
  dmScript.registerSiteSwitches()
  Script.registerSwitch('a', "All", "  Also show inactive replicas")
  Script.registerSwitch('', 'DiskOnly', '  Show only disk replicas')
  Script.registerSwitch('', 'PreferDisk', "  If disk replica, don't show tape replicas")
  Script.registerSwitch('', 'ForJobs', '  Select only replicas that can be used for jobs')
  Script.registerSwitch('', 'Protocol=',
                        '   Define the protocol for which a tURL is requested (default:root)')
  Script.registerSwitch('', 'Metalink',
                        '   Generate metalink files for parallel download with xrdcp \
                            (one file per lfn). Implies Protocol=root')
  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [option|cfgfile] ... [LFN[,LFN2[,LFN3...]]] SE[,SE2...]' % Script.scriptName,
                                    'Arguments:',
                                    '  LFN:      Logical File Name or file containing LFNs',
                                    '  SE:       Valid DIRAC SE']))
  Script.parseCommandLine(ignoreErrors=True)

  from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeAccessURL
  from DIRAC import exit
  exit(executeAccessURL(dmScript))


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_Resources_Catalog_BookkeepingDBClient."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

import unittest
import mock

import six

from DIRAC import gLogger
import LHCbDIRAC.Resources.Catalog.BookkeepingDBClient as moduleTested

__RCSID__ = "$Id$"

################################################################################


class BookkeepingDBClientt_TestCase(unittest.TestCase):

  def setUp(self):
    """Setup."""
    gLogger.setLevel('DEBUG')

    # # Mock external libraries / modules not interesting for the unit test
    # mock_pathFinder = mock.Mock()
    # mock_pathFinder.getServiceURL.return_value = 'cookiesURL'
    # self.mock_pathFinder = mock_pathFinder

    # Add mocks to moduleTested
    # moduleTested.PathFinder = self.mock_pathFinder

    mock_RPC = mock.Mock()
    mock_RPC.addFiles.return_value = {'OK': True, 'Value': {
        'Successful': ['A'], 'Failed': ['B']
    }}
    mock_RPC.removeFiles.return_value = {'OK': True, 'Value': {
        'Successful': ['A'], 'Failed': ['B']
    }}
    mock_RPC.exists.return_value = {'OK': True, 'Value': {
        'A': 1, 'B': 2
    }}
    mock_RPC.getFileMetadata.return_value = {'OK': True, 'Value': {
        'Successful': {'A': {'FileSize': 1}, 'B': {'FileSize': 2}}
    }}
    # mock_RPC.removeMigratingFiles.return_value    = {'OK': True}
    # mock_RPC.removeMigratingReplicas.return_value = {'OK': True}

    mock_BookkeepingClient = mock.Mock()
    mock_BookkeepingClient.return_value = mock_RPC
    self.mock_BookkeepingClient = mock_BookkeepingClient

    moduleTested.BookkeepingClient = self.mock_BookkeepingClient

    self.moduleTested = moduleTested
    self.testClass = self.moduleTested.BookkeepingDBClient

  def tearDown(self):
    """TearDown."""
    del self.testClass
    del self.moduleTested
    del self.mock_BookkeepingClient

################################################################################


class BookkeepingDBClient_Success(BookkeepingDBClientt_TestCase):

  def test_instantiate(self):
    """tests that we can instantiate one object of the tested class."""

    catalog = self.testClass()
    self.assertEqual('BookkeepingDBClient', catalog.__class__.__name__)

  def test_init(self):
    """tests that the init method does what it should do."""

    catalog = self.testClass()

    self.assertEqual(1000, catalog.splitSize)
    self.assertEqual('BookkeepingDB', catalog.name)

    catalog = self.testClass()
    self.assertEqual(1000, catalog.splitSize)
    self.assertEqual('BookkeepingDB', catalog.name)

  def test__setHasReplicaFlag(self):
    """test the output of __setHasReplicaFlag."""

    catalog = self.testClass()

    res = catalog._BookkeepingDBClient__setHasReplicaFlag([])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {}}, res['Value'])

    res = catalog._BookkeepingDBClient__setHasReplicaFlag(['A'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': "File does not exist"}}, res['Value'])

    res = catalog._BookkeepingDBClient__setHasReplicaFlag(['A', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': 'File does not exist'}}, res['Value'])

    res = catalog._BookkeepingDBClient__setHasReplicaFlag(['C'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': 'File does not exist'}}, res['Value'])

    mock_RPC = mock.Mock()
    mock_RPC.addFiles.return_value = {'OK': False, 'Message': 'Bo!'}

    self.moduleTested.BookkeepingClient.return_value = mock_RPC

    catalog = self.testClass()

    res = catalog._BookkeepingDBClient__setHasReplicaFlag([])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {}}, res['Value'])

    res = catalog._BookkeepingDBClient__setHasReplicaFlag(['A'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {'A': 'Bo!'}}, res['Value'])

    res = catalog._BookkeepingDBClient__setHasReplicaFlag(['A', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {'A': 'Bo!', 'B': 'Bo!'}}, res['Value'])

    res = catalog._BookkeepingDBClient__setHasReplicaFlag(['C'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {'C': 'Bo!'}}, res['Value'])

    mock_RPC = mock.Mock()
    # side_effect does not work very well, cooked a workaround
    _myValues = [{'OK': False, 'Message': 'Bo!'}, {'OK': True, 'Value': {'A': 1, 'B': 2}}]

    def _side_effect(_pfn):
      return _myValues.pop()
    mock_RPC.addFiles.side_effect = _side_effect
    # mock_RPC.addFiles.side_effect = [{'OK': True, 'Value': {'A': 1 , 'B': 2}},
    #                                  {'OK': False, 'Message': 'Bo!'}]

    self.moduleTested.BookkeepingClient.return_value = mock_RPC

    catalog = self.testClass()
    catalog.splitSize = 2

    res = catalog._BookkeepingDBClient__setHasReplicaFlag([])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {}}, res['Value'])

    _myValues = [{'OK': False, 'Message': 'Bo!'}, {'OK': True, 'Value': {'Successful': ['A'], 'Failed':[]}}]

    res = catalog._BookkeepingDBClient__setHasReplicaFlag(['A', 'C', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': 'Bo!'}}, res['Value'])

    # Restore the module
    self.moduleTested.BookkeepingClient.return_value = self.mock_BookkeepingClient
    six.moves.reload_module(self.moduleTested)

  def test__unsetHasReplicaFlag(self):
    """test the output of __unsetHasReplicaFlag."""

    catalog = self.testClass()

    res = catalog._BookkeepingDBClient__unsetHasReplicaFlag([])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {}}, res['Value'])

    res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(['A'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': "File does not exist"}}, res['Value'])

    res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(['A', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': 'File does not exist'}}, res['Value'])

    res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(['C'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': 'File does not exist'}}, res['Value'])

    mock_RPC = mock.Mock()
    mock_RPC.removeFiles.return_value = {'OK': False, 'Message': 'Bo!'}

    self.moduleTested.BookkeepingClient.return_value = mock_RPC

    catalog = self.testClass()

    res = catalog._BookkeepingDBClient__unsetHasReplicaFlag([])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {}}, res['Value'])

    res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(['A'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {'A': 'Bo!'}}, res['Value'])

    res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(['A', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {'A': 'Bo!', 'B': 'Bo!'}}, res['Value'])

    res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(['C'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {'C': 'Bo!'}}, res['Value'])

    mock_RPC = mock.Mock()
    # side_effect does not work very well, cooked a workaround
    _myValues = [{'OK': False, 'Message': 'Bo!'}, {'OK': True, 'Value': {'Successful': [], 'Failed':['A', 'B']}}]

    def _side_effect(_pfn):
      return _myValues.pop()
    mock_RPC.removeFiles.side_effect = _side_effect
    # mock_RPC.removeFiles.side_effect = [{'OK': True, 'Value': {'A': 1 , 'B': 2}},
    #                                     {'OK': False, 'Message': 'Bo!'}]

    self.moduleTested.BookkeepingClient.return_value = mock_RPC

    catalog = self.testClass()
    catalog.splitSize = 2

    res = catalog._BookkeepingDBClient__unsetHasReplicaFlag([])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {}}, res['Value'])

    res = catalog._BookkeepingDBClient__unsetHasReplicaFlag(['A', 'C', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {'A': 'File does not exist', 'B': 'Bo!'}}, res['Value'])

    # Restore the module
    self.moduleTested.BookkeepingClient.return_value = self.mock_BookkeepingClient
    six.moves.reload_module(self.moduleTested)

  def test__exists(self):
    """tests the output of __exists."""

    catalog = self.testClass()

    res = catalog._BookkeepingDBClient__exists([])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {}}, res['Value'])

    res = catalog._BookkeepingDBClient__exists(['A', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'B': 2}, 'Failed': {}}, res['Value'])

    res = catalog._BookkeepingDBClient__exists(['C'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'B': 2}, 'Failed': {}}, res['Value'])

    mock_RPC = mock.Mock()
    mock_RPC.exists.return_value = {'OK': False, 'Message': 'Bo!'}

    self.moduleTested.BookkeepingClient.return_value = mock_RPC

    catalog = self.testClass()

    res = catalog._BookkeepingDBClient__exists([])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {}}, res['Value'])

    res = catalog._BookkeepingDBClient__exists(['A'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {'A': 'Bo!'}}, res['Value'])

    res = catalog._BookkeepingDBClient__exists(['A', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {'A': 'Bo!', 'B': 'Bo!'}}, res['Value'])

    res = catalog._BookkeepingDBClient__exists(['C'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {'C': 'Bo!'}}, res['Value'])

    mock_RPC = mock.Mock()

    # side_effect does not work very well, cooked a workaround
    _myValues = [{'OK': False, 'Message': 'Bo!'}, {'OK': True, 'Value': {'A': 1, 'B': 2}}]

    def _side_effect(_pfn):
      return _myValues.pop()
    mock_RPC.exists.side_effect = _side_effect
    # mock_RPC.exists.side_effect = [{'OK': True, 'Value': {'A': 1 , 'B': 2}},
    #                                {'OK': False, 'Message': 'Bo!'}]

    self.moduleTested.BookkeepingClient.return_value = mock_RPC

    catalog = self.testClass()
    catalog.splitSize = 2

    res = catalog._BookkeepingDBClient__exists([])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {}}, res['Value'])

    _myValues = [{'OK': False, 'Message': 'Bo!'}, {'OK': True, 'Value': {'A': 1, 'B': 2}}]

    res = catalog._BookkeepingDBClient__exists(['A', 'C', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'B': 2}, 'Failed': {'B': 'Bo!'}}, res['Value'])

    # Restore the module
    self.moduleTested.BookkeepingClient.return_value = self.mock_BookkeepingClient
    six.moves.reload_module(self.moduleTested)

  def test__getFileMetadata(self):
    """tests the output of __getFileMetadata."""

    catalog = self.testClass()

    res = catalog._BookkeepingDBClient__getFileMetadata(['A', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': {'FileSize': 1}, 'B': {'FileSize': 2}}, 'Failed': {}}, res['Value'])

    res = catalog._BookkeepingDBClient__getFileMetadata(['A', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': {'FileSize': 1}, 'B': {'FileSize': 2}},
                      'Failed': {}},
                     res['Value'])

    res = catalog._BookkeepingDBClient__getFileMetadata(['C'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': {'FileSize': 1}, 'B': {'FileSize': 2}},
                      'Failed': {'C': 'File does not exist'}},
                     res['Value'])

    mock_RPC = mock.Mock()
    mock_RPC.getFileMetadata.return_value = {'OK': True, 'Value': {'Successful': {'A': '1', 'B': '2'}}}

    self.moduleTested.BookkeepingClient.return_value = mock_RPC

    catalog = self.testClass()

    res = catalog._BookkeepingDBClient__getFileMetadata(['A', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {'A': '1', 'B': '2'}}, res['Value'])

    res = catalog._BookkeepingDBClient__getFileMetadata(['C'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {},
                      'Failed': {'A': '1', 'B': '2', 'C': 'File does not exist'}},
                     res['Value'])

    mock_RPC = mock.Mock()
    mock_RPC.getFileMetadata.return_value = {'OK': False, 'Message': 'Bo!'}

    self.moduleTested.BookkeepingClient.return_value = mock_RPC

    catalog = self.testClass()

    res = catalog._BookkeepingDBClient__getFileMetadata(['A'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {'A': 'Bo!'}}, res['Value'])

    res = catalog._BookkeepingDBClient__getFileMetadata(['A', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {}, 'Failed': {'A': 'Bo!', 'B': 'Bo!'}}, res['Value'])

    mock_RPC = mock.Mock()

    # side_effect does not work very well, cooked a workaround
    _myValues = [{'OK': False, 'Message': 'Bo!'}, {'OK': True, 'Value': {'Successful': {'A': 1, 'B': 2}}}]

    def _side_effect(_pfn):
      return _myValues.pop()
    mock_RPC.getFileMetadata.side_effect = _side_effect
    # mock_RPC.getFileMetadata.side_effect = [{'OK': True, 'Value': {'A': 1 , 'B': 2}},
    #                                         {'OK': False, 'Message': 'Bo!'}]

    self.moduleTested.BookkeepingClient.return_value = mock_RPC

    catalog = self.testClass()
    catalog.splitSize = 2

    res = catalog._BookkeepingDBClient__getFileMetadata(['A', 'C', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'B': 2}, 'Failed': {'C': 'File does not exist',
                                                                 'B': 'Bo!'}}, res['Value'])

    mock_RPC = mock.Mock()
    mock_RPC.getFileMetadata.return_value = {'OK': True, 'Value': {'Successful': {'A': str, 'B': '2'}}}

    self.moduleTested.BookkeepingClient.return_value = mock_RPC

    catalog = self.testClass()

    res = catalog._BookkeepingDBClient__getFileMetadata(['A', 'B'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': str}, 'Failed': {'B': '2'}}, res['Value'])

    # Restore the module
    self.moduleTested.BookkeepingClient.return_value = self.mock_BookkeepingClient
    six.moves.reload_module(self.moduleTested)

  def test_addFile(self):
    """tests the output of addFile."""

    catalog = self.testClass()

    res = catalog.addFile(1)
    self.assertEqual(False, res['OK'])

    res = catalog.addFile([])
    self.assertEqual(False, res['OK'])

    res = catalog.addFile({})
    self.assertEqual(False, res['OK'])

    res = catalog.addFile(['path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': 'File does not exist'}}, res['Value'])

    res = catalog.addFile(['A'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': 'File does not exist'}}, res['Value'])

    res = catalog.addFile(['A', 'path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': 'File does not exist'}}, res['Value'])

  def test_addReplica(self):
    """tests the output of addReplica."""

    catalog = self.testClass()

    res = catalog.addReplica(1)
    self.assertEqual(False, res['OK'])

    res = catalog.addReplica([])
    self.assertEqual(False, res['OK'])

    res = catalog.addReplica({})
    self.assertEqual(False, res['OK'])

    res = catalog.addReplica(['path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': 'File does not exist'}}, res['Value'])

    res = catalog.addReplica(['A'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': 'File does not exist'}}, res['Value'])

    res = catalog.addReplica(['A', 'path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': 'File does not exist'}}, res['Value'])

  def test_removeFile(self):
    """tests the output of removeFile."""

    catalog = self.testClass()

    res = catalog.removeFile(1)
    self.assertEqual(False, res['OK'])

    res = catalog.removeFile([])
    self.assertEqual(False, res['OK'])

    res = catalog.removeFile({})
    self.assertEqual(False, res['OK'])

    res = catalog.removeFile(['A', 'B', 'C'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': 'File does not exist'}}, res['Value'])

    res = catalog.removeFile({'A': 1, 'B': 2, 'C': 3})
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': True}, 'Failed': {'B': 'File does not exist'}}, res['Value'])

    # FIXME: to be continued, but the method is quite tangled to be tested
    # on a rational way

  def test_removeReplica(self):
    """tests the output of removeReplica."""

    catalog = self.testClass()

    res = catalog.removeReplica(1)
    self.assertEqual(False, res['OK'])

    res = catalog.removeReplica([])
    self.assertEqual(False, res['OK'])

    res = catalog.removeReplica({})
    self.assertEqual(False, res['OK'])

    res = catalog.removeReplica(['path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True}, 'Failed': {}}, res['Value'])

    res = catalog.removeReplica(['A', 'B', 'path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True, 'A': True, 'B': True}, 'Failed': {}}, res['Value'])

    res = catalog.removeReplica({'A': 2, 'C': 3})
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'C': True}, 'Failed': {}}, res['Value'])

  def test_setReplicaStatus(self):
    """tests the output of setReplicaStatus."""

    catalog = self.testClass()

    res = catalog.setReplicaStatus(1)
    self.assertEqual(False, res['OK'])

    res = catalog.setReplicaStatus([])
    self.assertEqual(False, res['OK'])

    res = catalog.setReplicaStatus({})
    self.assertEqual(False, res['OK'])

    res = catalog.setReplicaStatus(['path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True}, 'Failed': {}}, res['Value'])

    res = catalog.setReplicaStatus(['A', 'B', 'path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True, 'A': True, 'B': True}, 'Failed': {}}, res['Value'])

    res = catalog.setReplicaStatus({'A': 2, 'C': 3})
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'C': True}, 'Failed': {}}, res['Value'])

  def test_setReplicaHost(self):
    """tests the output of setReplicaHost."""

    catalog = self.testClass()

    res = catalog.setReplicaHost(1)
    self.assertEqual(False, res['OK'])

    res = catalog.setReplicaHost([])
    self.assertEqual(False, res['OK'])

    res = catalog.setReplicaHost({})
    self.assertEqual(False, res['OK'])

    res = catalog.setReplicaHost(['path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True}, 'Failed': {}}, res['Value'])

    res = catalog.setReplicaHost(['A', 'B', 'path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True, 'A': True, 'B': True}, 'Failed': {}}, res['Value'])

    res = catalog.setReplicaHost({'A': 2, 'C': 3})
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'C': True}, 'Failed': {}}, res['Value'])

  def test_removeDirectory(self):
    """tests the output of removeDirectory."""

    catalog = self.testClass()

    res = catalog.removeDirectory(1)
    self.assertEqual(False, res['OK'])

    res = catalog.removeDirectory([])
    self.assertEqual(False, res['OK'])

    res = catalog.removeDirectory({})
    self.assertEqual(False, res['OK'])

    res = catalog.removeDirectory(['path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True}, 'Failed': {}}, res['Value'])

    res = catalog.removeDirectory(['A', 'B', 'path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True, 'A': True, 'B': True}, 'Failed': {}}, res['Value'])

    res = catalog.removeDirectory({'A': 2, 'C': 3})
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'C': True}, 'Failed': {}}, res['Value'])

  def test_createDirectory(self):
    """tests the output of createDirectory."""

    catalog = self.testClass()

    res = catalog.createDirectory(1)
    self.assertEqual(False, res['OK'])

    res = catalog.createDirectory([])
    self.assertEqual(False, res['OK'])

    res = catalog.createDirectory({})
    self.assertEqual(False, res['OK'])

    res = catalog.createDirectory(['path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True}, 'Failed': {}}, res['Value'])

    res = catalog.createDirectory(['A', 'B', 'path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True, 'A': True, 'B': True}, 'Failed': {}}, res['Value'])

    res = catalog.createDirectory({'A': 2, 'C': 3})
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'C': True}, 'Failed': {}}, res['Value'])

  def test_removeLink(self):
    """tests the output of createDirectory."""

    catalog = self.testClass()

    res = catalog.removeLink(1)
    self.assertEqual(False, res['OK'])

    res = catalog.removeLink([])
    self.assertEqual(False, res['OK'])

    res = catalog.removeLink({})
    self.assertEqual(False, res['OK'])

    res = catalog.removeLink(['path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True}, 'Failed': {}}, res['Value'])

    res = catalog.removeLink(['A', 'B', 'path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True, 'A': True, 'B': True}, 'Failed': {}}, res['Value'])

    res = catalog.removeLink({'A': 2, 'C': 3})
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'C': True}, 'Failed': {}}, res['Value'])

  def test_createLink(self):
    """tests the output of createDirectory."""

    catalog = self.testClass()

    res = catalog.createLink(1)
    self.assertEqual(False, res['OK'])

    res = catalog.createLink([])
    self.assertEqual(False, res['OK'])

    res = catalog.createLink({})
    self.assertEqual(False, res['OK'])

    res = catalog.createLink(['path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True}, 'Failed': {}}, res['Value'])

    res = catalog.createLink(['A', 'B', 'path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'path': True, 'A': True, 'B': True}, 'Failed': {}}, res['Value'])

    res = catalog.createLink({'A': 2, 'C': 3})
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'C': True}, 'Failed': {}}, res['Value'])

  def test_exists(self):
    """tests the output of exists."""

    catalog = self.testClass()

    res = catalog.exists(1)
    self.assertEqual(False, res['OK'])

    res = catalog.exists([])
    self.assertEqual(False, res['OK'])

    res = catalog.exists({})
    self.assertEqual(False, res['OK'])

    res = catalog.exists(['path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'B': 2}, 'Failed': {}}, res['Value'])

    res = catalog.exists(['A', 'B', 'path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'B': 2}, 'Failed': {}}, res['Value'])

    res = catalog.exists({'A': 2, 'C': 3})
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'B': 2}, 'Failed': {}}, res['Value'])

  def test_getFileMetadata(self):
    """tests the output of getFileMetadata."""

    catalog = self.testClass()

    res = catalog.getFileMetadata(1)
    self.assertEqual(False, res['OK'])

    res = catalog.getFileMetadata([])
    self.assertEqual(False, res['OK'])

    res = catalog.getFileMetadata({})
    self.assertEqual(False, res['OK'])

    res = catalog.getFileMetadata(['path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': {'FileSize': 1}, 'B': {'FileSize': 2}},
                      'Failed': {'path': 'File does not exist'}},
                     res['Value'])

    res = catalog.getFileMetadata(['A', 'B', 'path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': {'FileSize': 1}, 'B': {'FileSize': 2}},
                      'Failed': {'path': 'File does not exist'}},
                     res['Value'])

    res = catalog.getFileMetadata({'A': 2, 'C': 3})
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': {'FileSize': 1}, 'B': {'FileSize': 2}},
                      'Failed': {'C': 'File does not exist'}},
                     res['Value'])

  def test_getFileSize(self):
    """tests the output of getFileSize."""

    catalog = self.testClass()

    res = catalog.getFileSize(1)
    self.assertEqual(False, res['OK'])

    res = catalog.getFileSize([])
    self.assertEqual(False, res['OK'])

    res = catalog.getFileSize({})
    self.assertEqual(False, res['OK'])

    res = catalog.getFileSize(['path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'B': 2}, 'Failed': {'path': 'File does not exist'}}, res['Value'])

    res = catalog.getFileSize(['A', 'B', 'path'])
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'B': 2}, 'Failed': {'path': 'File does not exist'}}, res['Value'])

    res = catalog.getFileSize({'A': 2, 'C': 3})
    self.assertEqual(True, res['OK'])
    self.assertEqual({'Successful': {'A': 1, 'B': 2}, 'Failed': {'C': 'File does not exist'}}, res['Value'])

#################################################################################



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Uses the DM script switches, and, unless a list of LFNs is provided

1. If --BKQuery is used: get files in BK directories, check if they are in FC
2. If --Production is used get files using the bk query of the given production

Then check if files registered as having a replica in the BK are also in the FC.

If --FixIt is set, take actions

- add files to the BK if they exist in the FC, but have replica = NO in the BK
- set replicaFlag = No in the BK for those files that are not in the FC
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  # Script initialization
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, Script
  from DIRAC import gLogger

  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [option|cfgfile] [values]' % Script.scriptName, ]))
  dmScript = DMScript()
  dmScript.registerBKSwitches()
  dmScript.registerFileSwitches()
  Script.registerSwitch('', 'FixIt', '   Take action to fix the catalogs')
  Script.registerSwitch('', 'CheckAllFlags', '   Consider also files with replica flag NO')
  Script.parseCommandLine(ignoreErrors=True)

  fixIt = False
  checkAll = False
  for opt, val in Script.getUnprocessedSwitches():
    if opt == 'FixIt':
      fixIt = True
    elif opt == 'CheckAllFlags':
      checkAll = True

  # imports
  from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks
  gLogger.setLevel('INFO')
  cc = ConsistencyChecks()
  bkQuery = dmScript.getBKQuery(visible='All')
  cc.bkQuery = bkQuery
  cc.lfns = dmScript.getOption('LFNs', [])
  productions = dmScript.getOption('Productions', [])

  from LHCbDIRAC.DataManagementSystem.Client.CheckExecutors import doCheckBK2FC
  if productions:
    for prod in productions:
      cc.prod = prod
      gLogger.always("Processing production %d" % cc.prod)
      doCheckBK2FC(cc, checkAll, fixIt)
      gLogger.always("Processed production %d" % cc.prod)
  else:
    doCheckBK2FC(cc, checkAll, fixIt)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: TransformationCleaningAgent

.. module: TransformationCleaningAgent

:synopsis: clean up of finalised transformations
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import ast

import six

# # from DIRAC
from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.TransformationSystem.Agent.TransformationCleaningAgent import TransformationCleaningAgent as DiracTCAgent
# # from LHCbDIRAC
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient

# # agent's name
AGENT_NAME = 'Transformation/TransformationCleaningAgent'


class TransformationCleaningAgent(DiracTCAgent):
  """ .. class:: TransformationCleaningAgent
  """

  def __init__(self, *args, **kwargs):
    """c'tor."""
    DiracTCAgent.__init__(self, *args, **kwargs)

    self.directoryLocations = ['TransformationDB', 'StorageUsage']
    self.archiveAfter = 7
    self.fileTypesToKeep = ['GAUSSHIST']

    self.bkClient = None
    self.transClient = None
    self.storageUsageClient = None

  #############################################################################

  def initialize(self):
    """Standard initialize method for agents."""
    DiracTCAgent.initialize(self)

    self.directoryLocations = sorted(self.am_getOption('DirectoryLocations', self.directoryLocations))
    self.archiveAfter = self.am_getOption('ArchiveAfter', self.archiveAfter)  # days

    self.fileTypesToKeep = Operations().getValue('Transformations/FileTypesToKeep', self.fileTypesToKeep)

    self.bkClient = BookkeepingClient()
    self.transClient = TransformationClient()
    self.storageUsageClient = StorageUsageClient()

    return S_OK()

  def cleanMetadataCatalogFiles(self, transID):
    """clean the metadata using BKK and Data Manager. This method is a
    replacement of the one from base class.

    :param self: self reference
    :param int transID: transformation ID
    """
    res = self.bkClient.getProductionFiles(transID, 'ALL', 'Yes')
    if not res['OK']:
      return res
    bkMetadata = res['Value']
    fileToRemove = []
    yesReplica = []
    self.log.info("Found a total of %d files in the BK for transformation %d" % (len(bkMetadata), transID))
    for lfn, metadata in bkMetadata.items():
      if metadata['FileType'] != 'LOG':
        fileToRemove.append(lfn)
        if metadata['GotReplica'] == 'Yes':
          yesReplica.append(lfn)
    if fileToRemove:
      self.log.info("Attempting to remove %d possible remnants from the catalog and storage" % len(fileToRemove))
      # Executing with shifter proxy
      gConfigurationData.setOptionInCFG('/DIRAC/Security/UseServerCertificate', 'false')
      res = DataManager().removeFile(fileToRemove, force=True)
      gConfigurationData.setOptionInCFG('/DIRAC/Security/UseServerCertificate', 'true')
      if not res['OK']:
        return res
      for lfn, reason in res['Value']['Failed'].items():
        self.log.error("Failed to remove file found in BK", "%s %s" % (lfn, reason))
      if res['Value']['Failed']:
        return S_ERROR("Failed to remove all files found in the BK")
      if yesReplica:
        self.log.info("Ensuring that %d files are removed from the BK" % (len(yesReplica)))
        res = FileCatalog(catalogs=['BookkeepingDB']).removeFile(yesReplica)
        if not res['OK']:
          return res
        for lfn, reason in res['Value']['Failed'].items():
          self.log.error("Failed to remove file from BK", "%s %s" % (lfn, reason))
        if res['Value']['Failed']:
          return S_ERROR("Failed to remove all files from the BK")
    self.log.info("Successfully removed all files found in the BK")
    return S_OK()

  def getTransformationDirectories(self, transID):
    """get the directories for the supplied transformation from the
    transformation system.

    :param self: self reference
    :param int transID: transformation ID
    """

    res = DiracTCAgent.getTransformationDirectories(self, transID)

    if not res['OK']:
      return res

    directories = res['Value']
    if isinstance(directories, six.string_types):  # Check for (stupid) formats
      directories = ast.literal_eval(directories)
      if not isinstance(directories, list):
        return S_ERROR("Wrong format of output directories")

    if 'StorageUsage' in self.directoryLocations:
      res = self.storageUsageClient.getStorageDirectories('', '', transID, [])
      if not res['OK']:
        self.log.error("Failed to obtain storage usage directories", res['Message'])
        return res
      transDirectories = res['Value']
      directories = self._addDirs(transID, transDirectories, directories)

    if not directories:
      self.log.info("No output directories found")

    # We should be removing from the list of directories
    # those directories created for file types that are part of those:
    # - uploaded (as output files)
    # - not merged by subsequent steps
    # but this is pretty difficult to identify at run time, so we better remove the "RemovingFiles" production status
    # and replace it with a flush (this applies only to MC).
    # So we just have a created list.
    fileTypesToKeepDirs = []
    for fileTypeToKeep in self.fileTypesToKeep:
      fileTypesToKeepDirs.extend([x for x in directories if fileTypeToKeep in x])
    directories = list(set(directories).difference(set(fileTypesToKeepDirs)))

    directories = sorted(directories)
    return S_OK(directories)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Check if all files have a replica in a certain (set of) SE )Tier1-Archive
default) List the files that don't have a replica in the specified SE (group)"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  # Script initialization
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, Script

  Script.registerSwitch('', 'FixIt', '   Take action to fix the catalogs')
  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [option|cfgfile] [values]' % Script.scriptName, ]))
  dmScript = DMScript()
  dmScript.registerDMSwitches()  # Directory
  Script.parseCommandLine(ignoreErrors=True)
  fixIt = False
  for opt, val in Script.getUnprocessedSwitches():
    if opt == 'FixIt':
      fixIt = True

  # imports
  from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks
  cc = ConsistencyChecks()
  cc.directories = dmScript.getOption('Directory', [])
  cc.lfns = dmScript.getOption('LFNs', []) + [lfn for arg in Script.getPositionalArgs() for lfn in arg.split(',')]
  bkQuery = dmScript.getBKQuery(visible='All')
  if bkQuery.getQueryDict() != {'Visible': 'All'}:
    bkQuery.setOption('ReplicaFlag', 'All')
    cc.bkQuery = bkQuery
  seList = dmScript.getOption('SEs', [])
  if not seList:
    dmScript.setSEs('Tier1-Archive')
    seList = dmScript.getOption('SEs', [])

  from LHCbDIRAC.DataManagementSystem.Client.CheckExecutors import doCheckSE
  doCheckSE(cc, seList, fixIt)


if __name__ == "__main__":
  main()
